import time
import json
import openai

from .utils.prompt_base import EDITOR_PROMPT
from .utils.api import chatgpt

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver


@Solver.register("chatgpt_post_editor", "claim_info", "claim_info")
class ChatGPTPostEditor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.path_save_edited_claims = args.get("path_save_edited_claims", "evidence_stance_edit.json")
        self.num_retries = args.get("num_retries", 3)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)

        for key, pair in claim_info.items():
            claim = pair['claim'].strip()
            evids = pair['evidence_list']
            stance_explanation = pair['stances']
            # For not checkworthy claims, there is not a list of evidence (stances)
            if len(evids) == 0:
                claim_info[key].set({"edited_claims": claim, "operation": "no-check, no-edit"})
                continue

            # For checkworthy claims, with a list of evidence (stances)
            stance_label = [s.split()[0][:-1].lower() for s in stance_explanation]
            # print(key, stance_label)
            # rules to determine whether and how to edit:
            # if there is one support among stances, claim is regarded as true
            if "support" in stance_label:
                claim_info[key]['edited_claims'] = claim
                claim_info[key]['operation'] = "true, no-edit"
                # claim_info[key].set({"edited_claims": claim, "operation": "true, no-edit"})
            # if all stances are other, not direct/relevant to refute/support the claim, we delete it
            elif all([True for l in stance_label if l == "other"]):
                claim_info[key]['edited_claims'] = ''
                claim_info[key]['operation'] = "no-relevant-evidence, delete"
                # claim_info[key].set({"edited_claims": '', "operation": "no-relevant-evidence, delete"})
            # deal with refute-label with not mention explanation, these evidence is similar to other, just label is ambiguous
            elif all([True for l in stance_explanation if ("other," in l or " not mention" in l)]):
                claim_info[key]['edited_claims'] = ''
                claim_info[key]['operation'] = "no-relevant-evidence, delete"
                # claim_info[key].set({"edited_claims": '', "operation": "no-relevant-evidence, delete"})
            else:
                for i, s in enumerate(stance_label):
                    if s == "other":
                        continue
                    elif s == "refute" and "not mention" in stance_explanation[i]:
                        continue
                    elif s == "refute" or s == "partial support":
                        user_input = EDITOR_PROMPT.format(claim, evids[i])
                        for _ in range(self.num_retries):
                            try:
                                edits = chatgpt(user_input)
                                break
                            except openai.OpenAIError as exception:
                                print(f"{exception}. Retrying...")
                                time.sleep(1)
                        # update claim to revised claim as well
                        claim = edits
                        claim_info[key]["edited_claims"] = edits
                        claim_info[key]["operation"] = "false (refute or partial support), edit"
                        # claim_info[key].set(
                        #     {"edited_claims": edits, "operation": "false (refute or partial support), edit"})
                    else:
                        print(claim)
                        print(s, evids[i])

                        # write to json file
        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(self.path_save_edited_claims, "w") as outfile:
            outfile.write(json_object)

        state.set(self.output_name, claim_info)
        return True, state



import random

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("confused_claim_examiner", "claims_with_evidences", "claims_with_tags")
class ConfusedClaimExaminer(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        claims_with_tags = {}
        for claim_key, docs in claims.items():
            claims_with_tags[claim_key] = random.choice([True, False])

        state.set(self.output_name, claims_with_tags)
        return True, state



import random

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("fake_claim_extractor", "response", "claims")
class FakeClaimExtractor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_claims = args.get("max_claims", 5)
        self.min_claims = args.get("min_claims", 2)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        response = state.get(self.input_name)

        response_len = len(response)
        num_claims = random.randint(min(self.min_claims, response_len), min(self.min_claims, response_len))
        cut_pont = list(range(response_len))
        random.shuffle(cut_pont)
        cut_pont = sorted(cut_pont[:num_claims + 1])
        claims = []
        for i in range(len(cut_pont) - 1):
            claims.append(response[cut_pont[i]:cut_pont[i + 1]])

        state.set(self.output_name, claims)
        return True, state



"""Prompts for generating hallucinations."""

EVIDENCE_HALLUCINATION = """Generate a paragraph that answers the question.

Question: What is New York-Style pizza?
Text: New York-style pizza has slices that are large and wide with a thin crust that is foldable yet crispy. It is traditionally topped with tomato sauce and mozzarella cheese, with any extra toppings placed on top of the cheese.

Question: When did the first McDonald's open?
Text: The McDonald's brothers opened their first McDonald's restaurant in 1940 in San Bernardino, California. Originally, a carhop drive-in system was used to serve customers. The initial menu items were centered around barbecue and the first name the brothers called their business was "McDonald's Famous Barbecue."

Question: {query}
Text:
""".strip()



import os
import re
import openai
from openai import OpenAI
import requests
from typing import Any, Dict, List, Tuple

# ----------------------------------------------------------
# OpenAI ChatGPT and davicci-text
# ----------------------------------------------------------
client = None


def init_client():
    global client
    if client is None:
        if openai.api_key is None and "OPENAI_API_KEY" not in os.environ:
            print("openai_key not presented, delay to initialize.")
            return
        client = OpenAI()


def chatgpt(user_input):
    response = client.chat.completions.create(
        model="gpt-4o",
        messages=[
            {"role": "system", "content": "You are a NLP expert that is good at fact checking"},
            {"role": "user", "content": user_input},
        ],
    )

    result = ""
    for choice in response.choices:
        result += choice.message.content

    return result


def davinci(prompt):
    # Set up the model and prompt
    model_engine = "gpt-4o-instruct"

    # Generate a response
    completion = client.completions.create(
        engine=model_engine,
        prompt=prompt,
        max_tokens=1024,
        n=1,
        stop=None,
        temperature=0.5,
    )

    response = completion.choices[0].text
    return response


# ----------------------------------------------------------
# Bing Search
# ----------------------------------------------------------
BING_SEARCH_URL = "https://api.bing.microsoft.com/v7.0/search/"
SUBSCRIPTION_KEY = ""  # fill your bing api key


def search_bing(query: str, timeout: float = 3) -> List[str]:
    """Searches the query using Bing.
    Args:
        query: Search query.
        timeout: Timeout of the requests call.
    Returns:
        search_results: A list of the top URLs relevant to the query.
    """

    headers = {"Ocp-Apim-Subscription-Key": SUBSCRIPTION_KEY}
    params = {"q": query, "textDecorations": True, "textFormat": "HTML"}
    response = requests.get(BING_SEARCH_URL, headers=headers, params=params, timeout=timeout)
    response.raise_for_status()

    response = response.json()
    search_results = [r["url"] for r in response["webPages"]["value"]]
    return search_results


# Test Bing search
# search_results = search_bing("What are the different awards that Preslav Nakov has received")
# print(search_results)


# ----------------------------------------------------------
# Google Search
# ----------------------------------------------------------
def search_google(query: str, num_web_pages: int = 10, save_url: str = "") -> List[str]:
    """Searches the query using Google.
    Args:
        query: Search query.
        num_web_pages: the number of web pages to request.
        save_url: path to save returned urls, such as 'urls.txt'
    Returns:
        search_results: A list of the top URLs relevant to the query.
    """
    query = query.replace(" ", "+")

    # set headers: Google returns different web-pages according to agent device
    # desktop user-agent
    USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:65.0) Gecko/20100101 Firefox/65.0"
    # mobile user-agent
    MOBILE_USER_AGENT = "Mozilla/5.0 (Linux; Android 7.0; SM-G930V Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.125 Mobile Safari/537.36"
    headers = {"User-Agent": USER_AGENT}

    # set language
    # set the Google interface language, use &hl=XX
    # set the preferred language of the search results, use &lr=lang_XX
    # set language as en, otherwise it will return many translation web pages to Arabic that can't be opened correctly.
    lang = "en"

    # scrape google results
    urls = []
    for page in range(0, num_web_pages, 10):
        # here page is google search's bottom page meaning, click 2 -> start=10
        # url = "https://www.google.com/search?q={}&start={}".format(query, page)
        url = "https://www.google.com/search?q={}&lr=lang_{}&hl={}&start={}".format(query, lang, lang, page)
        r = requests.get(url, headers=headers)
        # collect all urls by regular expression
        # how to do if I just want to have the returned top-k pages?
        urls += re.findall('href="(https?://.*?)"', r.text)

    # set to remove repeated urls
    urls = list(set(urls))

    # save all url into a txt file
    if not save_url == "":
        with open(save_url, "w") as file:
            for url in urls:
                file.write(url + "\n")
    return urls


# Test google search
# query = "Google Company Introduction"
# urls = search_google(query)
# print(len(urls))



from .rarr_utils import search

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("search_engine_retriever", "claims_with_questions", "claims_with_evidences")
class RARRSearchEngineRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.max_search_results_per_query = args.get("max_search_results_per_query", 5)
        self.max_sentences_per_passage = args.get("max_sentences_per_passage", 4)
        self.sliding_distance = args.get("sliding_distance", 1)
        self.max_passages_per_search_result = args.get("max_passages_per_search_result", 1)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)

        for claim, contents in claims.items():
            questions = contents.get("questions", [])
            evidences = []
            for question in questions:
                evidences.extend(
                    search.run_search(
                        query=question,
                        max_search_results_per_query=self.max_search_results_per_query,
                        max_sentences_per_passage=self.max_sentences_per_passage,
                        sliding_distance=self.sliding_distance,
                        max_passages_per_search_result_to_return=self.max_passages_per_search_result,
                    )
                )
            claims[claim]['evidences'] = evidences

        state.set(self.output_name, claims)
        return True, state



import os
import importlib

from openfactcheck.utils.logging import get_logger
from openfactcheck.state import FactCheckerState

# Get the logger
logger = get_logger(__name__)

# Global solver registry
SOLVER_REGISTRY = {}

class StandardTaskSolver:
    """
    A class to represent a standard task solver. A standard task solver is a
    class that implements a specific task in a fact-checking system. It
    receives a FactCheckerState object as input and returns a new
    FactCheckerState object as output.

    Parameters
    ----------
    args : dict
        A dictionary containing the arguments to be passed to the solver.
    """
    
    name: str = None
    input_name: str = None
    output_name: str = None
    global_config: dict = dict()

    def __init__(self, args: dict):
        self.logger = logger
        self.args = args

        logger.debug(self.args)

    def __call__(self, state: FactCheckerState, **kwargs) -> tuple[
        bool, FactCheckerState]:
        raise NotImplementedError

    @classmethod
    def build_solver(cls, args):
        raise NotImplementedError

    @property
    def input_name(self):
        return self.__class__.input_name

    @property
    def output_name(self):
        return self.__class__.output_name

    def __str__(self):
        return f'[name:"{self.__class__.name}", input: "{self.__class__.input_name}": output: "{self.__class__.output_name}"]'

class Solver:
    """
    Class to handle the registration and loading of solvers
    """
    def __init__(self):
        pass

    def register(name, input_name=None, output_name=None):
        def decorator(cls):
            """
            Decorator to register a solver class
            """

            # Check if the solver is already registered
            if name in SOLVER_REGISTRY:
                return SOLVER_REGISTRY[name]

            # Check if the solver class extends StandardTaskSolver
            if not issubclass(cls, StandardTaskSolver):
                logger.error(f"Solver '{name}' must extend StandardTaskSolver, got {cls.__name__}.")
                raise ValueError(f"Solver '{name}' must extend StandardTaskSolver, got {cls.__name__}.")
            
            # Register the solver
            SOLVER_REGISTRY[name] = cls
            cls.name = name
            cls.input_name = input_name
            cls.output_name = output_name

            logger.info(f"Solver '{name}' registered")
            return cls

        return decorator
    
    @staticmethod
    def load_from_directory(directory, namespace):
        """
        Load solvers from a directory
        """

        # Check if the directory exists
        for item in sorted(os.listdir(directory), 
                           key=lambda x: os.path.isdir(os.path.join(directory, x)), 
                           reverse=True):
            
            # Skip hidden files and directories
            if item.startswith('_') or item.startswith('.'):
                continue

            # Get the full path of the item
            full_path = os.path.join(directory, item)

            # Load the item
            if os.path.isdir(full_path):
                Solver.load_from_directory(full_path, namespace + '.' + item)
            else:
                Solver.load_from_file(full_path, namespace)
    
    @staticmethod
    def load_from_file(file_path, namespace):
        """
        Load a solver from a file
        """

        # Check if the file is a Python file
        if file_path.endswith(".py"):
            # Get the solver name
            solver_name = os.path.basename(file_path)[:-3]

            # Get the module name
            module_name = namespace + "." + solver_name

            # Log the full module name to debug
            logger.debug(f"Attempting to import {module_name} from {file_path}")

            # Import the module
            try:
                importlib.import_module(module_name)
                logger.debug(f"Successfully imported {module_name}")
            except Exception as e:
                logger.error(f"Failed to import {module_name}: {e}")
                raise Exception(f"Failed to import {module_name}: {e}")

            return module_name

    @staticmethod
    def load(path, namespace):
        if os.path.isdir(path):
            Solver.load_from_directory(path, namespace)
        else:
            Solver.load_from_file(path, namespace)
        return



import io
import pandas as pd
from typing import Union
from importlib import resources as pkg_resources
from sklearn.metrics import accuracy_score, precision_score, f1_score, recall_score
from sklearn.metrics import classification_report, confusion_matrix

from openfactcheck import OpenFactCheck
from openfactcheck.templates import factchecker as templates_dir

# Import solver configuration templates
gold_claims_template_path = str(pkg_resources.files(templates_dir) / "gold/claims.jsonl")
gold_documents_template_path = str(pkg_resources.files(templates_dir) / "gold/documents.jsonl")

class CheckerEvaluator():
    """
    This class is used to evaluate the performance of a FactChecker.

    Parameters
    ----------
    input_path : Union[str, pd.DataFrame]
        The path to the CSV file or the DataFrame containing the FactChecker responses.
        The CSV file should have the following three columns:
        - label: The label assigned by the FactChecker. This should be a boolean value.
        - time: The time taken by the FactChecker to respond.
        - cost: The cost of the FactChecker response.
    eval_type : str
        The type of evaluation to perform. Either "claim" or "document".
    gold_path : str
        Optional. The path to the gold standard file. If not provided, the default gold standard file will be used.
        This is useful when evaluating the FactChecker on a different dataset.
    eval_type : str

    Attributes
    ----------
    input_path : Union[str, pd.DataFrame]
        The path to the CSV file or the DataFrame containing the FactChecker responses.
    gold_path : str
        The path to the gold standard file.
    eval_type : str
        The type of evaluation to perform. Either "claim" or "document".
    results : dict
        The evaluation results.
    confusion_matrix : numpy.ndarray
        The confusion matrix of the evaluation.
    classification_report : dict
        The classification report of the evaluation.

    Methods
    -------
    evaluate(input_path: Union[str, pd.DataFrame], eval_type: str, gold_path: str = ""):
        This function evaluates the performance of the FactChecker.
    evaluate_binary_classification(y_true, y_pred, pos_label="yes"):
        Evaluate the performance of a binary classification task.
    """
    def __init__(self, ofc: OpenFactCheck):
        """
        Initialize the FactCheckerEvaluator object.
        """
            
        # Set the attributes
        self.input_path = None
        self.gold_path = None
        self.eval_type = None
        self.results = None
        self.confusion_matrix = None
        self.classification_report = None

    @staticmethod
    def evaluate_binary_classification(y_true, y_pred, pos_label="yes"):
        accuracy = accuracy_score(y_true, y_pred)
        precision = precision_score(y_true, y_pred, pos_label=pos_label)
        recall = recall_score(y_true, y_pred, pos_label=pos_label)
        F1 = f1_score(y_true, y_pred, pos_label=pos_label)

        metrics = {
            "accuracy": round(accuracy, 3),
            "precision": round(precision, 3),
            "recall": round(recall, 3),
            "F1": round(F1, 3),
        }
        return metrics

    def evaluate(self, input_path: Union[str, pd.DataFrame], eval_type: str, gold_path: str = ""):
        """
        This function evaluates the performance of the FactChecker.
        """
        # Set the input_path, gold_path, and eval_type attributes
        self.input_path = input_path
        self.gold_path = gold_path
        self.eval_type = eval_type

        if self.gold_path == "":
            if eval_type == "claims":
                self.gold_path = gold_claims_template_path
            elif eval_type == "documents":
                self.gold_path = gold_documents_template_path
            else:
                raise ValueError("Invalid evaluation type. Please provide a valid evaluation type.")

        # Load the gold standard file
        with open(self.gold_path, "r") as f:
            json_data = f.read()
        df_gold = pd.read_json(io.StringIO(json_data), lines=True)

        # Check if the input_path is a DataFrame
        if isinstance(self.input_path, pd.DataFrame):
            df_input = self.input_path
        else:
            # Read the CSV file
            df_input = pd.read_csv(self.input_path)

        # Check if the FactChecker responses have the correct number of columns
        assert len(df_input.columns) == 3
        
        # Check if the FactChecker responses have the correct column names
        assert df_input.columns[0] == "label", f"The first column should be 'label' but is {df_input.columns[0]}."
        assert df_input.columns[1] == "time", f"The second column should be 'time' but is {df_input.columns[1]}."
        assert df_input.columns[2] == "cost", f"The third column should be 'cost' but is {df_input.columns[2]}."
        
        # Get the gold labels and the predictions
        if self.eval_type == "claims":
            gold_labels = df_gold['claim_label'].to_list()
        elif self.eval_type == "documents":
            gold_labels = df_gold['response_label'].to_list()
        predictions = df_input[df_input.columns[0]].to_list()

        # Check if the number of gold labels and predictions are the same
        assert (len(gold_labels) == len(predictions)), "The number of gold labels and predictions should be the same."

        # Verify that the gold labels and predictions are boolean values
        assert all(isinstance(label, bool) for label in gold_labels), "The gold labels should be boolean values."
        assert all(isinstance(label, bool) for label in predictions), "The predictions should be boolean values."

        # evalaute performance
        r1 = self.evaluate_binary_classification(y_true=gold_labels, y_pred=predictions, pos_label=True)
        r2 = self.evaluate_binary_classification(y_true=gold_labels, y_pred=predictions, pos_label=False)

        # Calculate total time and cost
        total_time = 0
        total_cost = 0
        
        # Check if the time and cost columns are present in the FactChecker responses
        if "time" in df_input.columns[1]:
            total_time = df_input[df_input.columns[1]].astype(float).sum()
        
        # Check if the cost column is present in the FactChecker responses
        if "cost" in df_input.columns[2]:
            total_cost = df_input[df_input.columns[2]].astype(float).sum()

        self.results = {
            "True_as_positive": r1,
            "False_as_positive": r2,
            "total_time": total_time,
            "total_cost": total_cost,
            "num_samples": len(predictions)
        }

        # Calculate the confusion matrix
        self.confusion_matrix = confusion_matrix(y_true=gold_labels, y_pred=predictions, labels=[True, False])

        # Calculate the classification report
        self.classification_report = classification_report(gold_labels, predictions)

        return self.results


import spacy
import logging

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("spacy_response_decomposer", 'response', 'sentences')
class SpacyResponseDecomposer(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        spacy_model = args.get("spacy_model", "en_core_web_sm")
        self.spacy_processor = spacy.load(spacy_model)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        response = state.get(self.input_name)
        doc = self.spacy_processor(response)
        sentences = [str(sent).strip() for sent in doc.sents]
        logging.info("The document is split into {} sentences.".format(len(sentences)))
        state.set(self.output_name, sentences)
        return True, state



import json

from .factool_utils.chat_api import OpenAIChat
from .factool_utils.prompt import VERIFICATION_PROMPT

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("factool_claim_examiner", "evidences", "claim_info")
class FactoolClaimExaminer(StandardTaskSolver):
    """
    A solver to examine the claims in a response.
    """
    def __init__(self, args):
        super().__init__(args)
        self.model_name = self.global_config.get("model_name", "gpt-4o")
        self.path_save_stance = args.get("path_save_stance", "evidence_stance.json")
        self.verifications = None
        self.gpt = OpenAIChat(self.model_name)
        self.verification_prompt = VERIFICATION_PROMPT

    # async def coro (self, factool_instance, claims_in_response, evidences):
    #    self.verifications = await factool_instance.pipelines["kbqa_online"]._verification(claims_in_response, evidences)
    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claim_info = state.get(self.input_name)
        # Recover the Factool objects
        claims_in_response = []
        queires = []
        search_outputs_for_claims = []
        for key, pair in claim_info.items():
            claim = key or pair["claim"]
            claims_in_response.append({"claim": claim})
            queires.append(pair["automatic_queries"])
            search_outputs_for_claim = []
            for evidence in pair["evidence_list"]:
                search_outputs_for_claim.append(
                    {
                        "content": evidence["web_page_snippet_manual"],
                        "source": evidence["url"],
                    }
                )
            search_outputs_for_claims.append(search_outputs_for_claim)

        claims_with_evidences = {k: [u['web_page_snippet_manual'] for u in claim_info[k]['evidence_list']] for k in
                                 claim_info.keys()}
        verifications = self._verification(claims_with_evidences)

        # evidences = [
        #     [output["content"] for output in search_outputs_for_claim]
        #     for search_outputs_for_claim in search_outputs_for_claims
        # ]

        # Attach the verifications (stances) to the claim_info
        for index, (key, pair) in enumerate(claim_info.items()):
            # print(f'Verifications: {verifications}\n')
            # print(f'Verification for claim {key}: Index {index}\n')
            # print(f'Verification for claim {key}: {verifications[index]}\n')
            # print(f'Verification for claim {key}: Type = {type(verifications[index])}\n')
            stance = ""
            index = 0  # Ensure the 'index' variable is defined somewhere appropriate in your context

            # Check if verifications at the current index is None or 'None'
            if verifications[index] is None or verifications[index] == "None":
                stance = claims_in_response[index]["claim"]
            else:
                # Initialize stance with error or empty string
                error = verifications[index].get("error", "")
                if error and error != "None":
                    stance = error + " "
                
                # Append reasoning if it exists and is not 'None'
                reasoning = verifications[index].get("reasoning", "")
                if reasoning and reasoning != "None":
                    stance += reasoning
                
                # Append claim or correction if available
                correction = verifications[index].get("correction", "")
                if correction and correction != "None":
                    stance += " " + correction
                else:
                    stance += claims_in_response[index]["claim"]
            claim_info[key]["stances"] = [stance]
            for j in range(len(claim_info[key]["evidence_list"])):
                claim_info[key]["evidence_list"][j]["stance"] = stance

        # write to json file
        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(self.path_save_stance, "w") as outfile:
            outfile.write(json_object)

        state.set(self.output_name, claim_info)
        return True, state

    def _verification(self, claims_with_evidences):
        messages_list = [
            [
                {"role": "system", "content": self.verification_prompt['system']},
                {"role": "user", "content": self.verification_prompt['user'].format(
                    claim=claim,
                    evidence=str([e[1] for e in evidence if isinstance(e, (list, tuple)) and len(e) > 1])
                )}
            ]
            for claim, evidence in claims_with_evidences.items()
        ]
        return self.gpt.run(messages_list, dict)



# code for general evaluation

import numpy as np
import evaluate
from sklearn.metrics import precision_recall_fscore_support, accuracy_score

def evaluate_classification(preds, gold):
    metric = evaluate.load("bstrai/classification_report")
    return metric.compute(predictions=preds, references=gold)

def eval_classification(y_true, y_pred, average="macro"):
    precision, recall, F1, support = precision_recall_fscore_support(y_true, y_pred, average=average)
    accuracy = accuracy_score(y_true, y_pred)

    metrics = {
        "accuracy": round(accuracy, 3),
        "precision": round(precision, 3),
        "recall": round(recall, 3),
        "F1": round(F1, 3),
    }
    return metrics


def eval_binary(y_true, y_pred, pos_label=1, average="binary"):
    """pos_label: postive label is machine text here, label is 1, human text is 0"""
    precision, recall, F1, support = precision_recall_fscore_support(
        y_true, y_pred, pos_label = pos_label, average = average)
    # accuracy
    accuracy = accuracy_score(y_true, y_pred)
    # precison
    # pre = precision_score(y_true, y_pred, pos_label = pos_label, average = average)
    # recall
    # rec = recall_score(y_true, y_pred, pos_label = pos_label, average = average)
    metrics = {
        "accuracy": round(accuracy, 3),
        "precision": round(precision, 3),
        "recall": round(recall, 3),
        "F1": round(F1, 3),
    }
    return metrics




from .utils.prompt_base import DECONTEXTILISATION_PROMPT
from .utils.api import chatgpt

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("chatgpt_decontextualizer", "sentences", "claims")
class ChatGPTDecontextualizer(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        sentences = state.get(self.input_name)

        if sentences is None:
            raise ValueError(f"sentences is required for {self}")

        results = []
        for sentence in sentences:
            user_input = DECONTEXTILISATION_PROMPT + sentence
            decontextualised_claims = chatgpt(user_input)
            print(decontextualised_claims)

            decontextualised_claims = decontextualised_claims.split("\n")
            decontextualised_claims = [claim.strip() for claim in decontextualised_claims if not claim.strip() == ""]
            decontextualised_claims = decontextualised_claims[1:]  # skip 'Output:'
            print("{} decontextualised claims.".format(len(decontextualised_claims)))
            results.extend(decontextualised_claims)

        state.set(self.output_name, results)
        return True, state



"""Utils for running question generation."""
import os
import time
from typing import List

import openai

openai.api_key = os.getenv("OPENAI_API_KEY")


def parse_api_response(api_response: str) -> List[str]:
    """Extract questions from the GPT-3 API response.

    Our prompt returns questions as a string with the format of an ordered list.
    This function parses this response in a list of questions.

    Args:
        api_response: Question generation response from GPT-3.
    Returns:
        questions: A list of questions.
    """
    search_string = "I googled:"
    questions = []
    for question in api_response.split("\n"):
        # Remove the search string from each question
        if search_string not in question:
            continue
        question = question.split(search_string)[1].strip()
        questions.append(question)

    return questions


def run_rarr_question_generation(
    claim: str,
    model: str,
    prompt: str,
    temperature: float,
    num_rounds: int,
    context: str = None,
    num_retries: int = 5,
) -> List[str]:
    """Generates questions that interrogate the information in a claim.

    Given a piece of text (claim), we use GPT-3 to generate questions that question the
    information in the claim. We run num_rounds of sampling to get a diverse set of questions.

    Args:
        claim: Text to generate questions off of.
        model: Name of the OpenAI GPT-3 model to use.
        prompt: The prompt template to query GPT-3 with.
        temperature: Temperature to use for sampling questions. 0 represents greedy deconding.
        num_rounds: Number of times to sample questions.
    Returns:
        questions: A list of questions.
    """
    if context:
        gpt3_input = prompt.format(context=context, claim=claim).strip()
    else:
        gpt3_input = prompt.format(claim=claim).strip()

    questions = set()
    for _ in range(num_rounds):
        for _ in range(num_retries):
            try:
                response = openai.completions.create(
                    model=model,
                    prompt=gpt3_input,
                    temperature=temperature,
                    max_tokens=256,
                )
                cur_round_questions = parse_api_response(
                    response.choices[0].text.strip()
                )
                questions.update(cur_round_questions)
                break
            except openai.OpenAIError as exception:
                print(f"{exception}. Retrying...")
                time.sleep(1)

    questions = list(sorted(questions))
    return questions



import os
import time
import json
import math
import pandas as pd
import seaborn as sns
from hashlib import md5
import matplotlib.pyplot as plt

from openfactcheck import OpenFactCheck
from openfactcheck.utils.logging import get_logger

# Get the logger
logger = get_logger(__name__)

class FreeTextEvaluator():
    def __init__(self, ofc: OpenFactCheck):
        """
        Initialize the FreeTextEvaluator object.
        """

        self.logger = logger
        
        # Set the OpenFactCheck object
        self.ofc = ofc

    def calculate_price(self, num_claims, cost_openai=0.015, cost_serper=0.001):
        """
        Calculate the cost (in USD) of the API calls for the free-text experiment.
        2x API calls per claim
        
        Parameters
        ----------
        numClaims : int
            The number of claims in the free-text experiment.
        costOpenAI : float
            The cost of the OpenAI API call.
        costSerper : float
            The cost of the Serper API call.
        """
        return num_claims * 2 * (cost_openai + cost_serper)

    def sum_all_elements(self, obj: dict):
        """
        Sum all elements of an object.
        """
        ret = 0
        for k, v in obj.items():
            ret += v
        return ret

    def assess_freetext(self, output_path: str):
        """
        Assess the free-text experiment, i.e., the number and type of claims, this is, Exact Matching (EM).
        """

        # Initialize the return object
        claims = {
            "num_false_claims": 0,
            "num_mixed_claims": 0,
            "num_true_claims": 0,
            "num_undefined_claims": 0
        }
        path = output_path + '/evidence_stance.json'
        if not os.path.exists(path):
            return False
        df = pd.read_json(path, lines=False)
        dataobj = json.loads(df.to_json())

        # Assess the claims
        for k, v in dataobj.items():
            # If stance contains definitive or mixed, then it is false
            if "definitive" in v["stances"][0] or "mixed" in v["stances"][0]:
                claims["num_mixed_claims"] += 1
            elif "factual" in v["stances"][0] or "confirm" in v["stances"][0]:
                claims["num_true_claims"] += 1
            elif "error" in v["stances"][0] or "incorrect" in v["stances"][0] or "false" in v["stances"][0]:
                claims["num_false_claims"] += 1
            else:
                claims["num_undefined_claims"] += 1

        return claims
    
    def read_evaluations(self):
        """
        Read the evaluations from the output directory.
        """
        data = []
        for dirname in os.listdir(self.base_output_path):
            dirpath = os.path.join(self.base_output_path, dirname)
            if os.path.isdir(dirpath):
                if os.path.exists(os.path.join(dirpath, 'evaluation.json')):
                    with open(os.path.join(dirpath, 'evaluation.json'), 'r') as f:
                        data.append(json.load(f))
        return data
    
    def read_results(self, evaluations):
        """
        Read the results from the evaluations.
        """
        # Calculate the total cost and time
        (costs, time_costs, true_claims, false_claims, mixed_claims, undefined_claims, total_claims) = (0, 0, 0, 0, 0, 0, 0)
        for evaluation in evaluations:
            total_claims += 1

            # Calculate the costs
            costs += self.calculate_price(self.sum_all_elements(evaluation["claims"]))
            time_costs += evaluation["end"] - evaluation["start"]

            # Calculate the number of claims
            false_claims += evaluation["claims"]["num_false_claims"]
            mixed_claims += evaluation["claims"]["num_mixed_claims"]
            undefined_claims += evaluation["claims"]["num_undefined_claims"]
            if (evaluation["claims"]["num_false_claims"] + evaluation["claims"]["num_mixed_claims"]) == 0:
                true_claims += 1
        
        return{
            "Claims": total_claims,
            "True Claims": true_claims,
            "False Claims": false_claims,
            "Mixed Claims": mixed_claims,
            "Undefined Claims": undefined_claims,
            "Cost (USD)": costs,
            "Time (ms)": time_costs,
            "Percentage of True Responses": round(true_claims / total_claims if total_claims != 0 else 0, 3) * 100,
            "Percentage of False Responses": round(false_claims / total_claims if total_claims != 0 else 0, 3) * 100
        }
    
    def freetext_barplot(self, results, fig_path: str = "", save: bool = False):
        """
        Create a barplot for the free-text evaluation results, ensuring full row utilization.

        Parameters
        ----------
        results : dict
            The dictionary of results from the free-text evaluation.
        fig_path : str
            The path to save the figure.
        save : bool
            Whether to save the figure or not.
        """

        # Exclude "Claims" and prepare data
        metrics = list(next(iter(results.values())).keys())
        datasets = list(results.keys())

        # Prepare plot data and handle specific conversions
        plot_data = {}
        for metric in metrics:
            if metric == "Claims":
                continue
            if metric == "Time (s)":
                plot_data["Time (min)"] = [results[dataset][metric] / (1000 * 60) for dataset in datasets] 
            elif metric == "Percentage of True Responses":
                plot_data[metric] = [results[dataset][metric] for dataset in datasets]
            else:
                plot_data[metric] = [results[dataset][metric] for dataset in datasets]

        # Define the layout
        total_metrics = len(plot_data)
        ncols = 4  # Maximum number of columns per row
        nrows = (total_metrics + ncols - 1) // ncols  # Calculate the required number of rows

        # Creating subplots
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(20, 5 * nrows))
        fig.suptitle('Performance on Free-Text Dataset')

        # Flatten axes array if more than one row
        axes = axes.flatten() if nrows > 1 else [axes]

        # Generate each bar plot and deactivate unused axes
        for ax, (metric, values) in zip(axes[:total_metrics], plot_data.items()):
            bars = ax.bar(datasets, values, color=sns.color_palette("rocket", n_colors=len(datasets)))
            ax.set_title(metric)
            ax.set_xticks(range(len(datasets)))
            ax.set_xticklabels(datasets, rotation=45, ha="right")
            ax.set_ylabel(metric)
            
            # Annotate each bar with its value
            for bar in bars:
                yval = bar.get_height()
                ax.text(bar.get_x() + bar.get_width()/2, yval, round(yval, 2),
                        ha='center', va='bottom')
                
            # Set y-axis limits to accommodate annotations
            ax.set_ylim(0, max(values) * 1.1)

        # Hide unused axes
        for ax in axes[total_metrics:]:
            ax.axis('off')

        # Adjust layout to prevent overlap
        plt.tight_layout()

        if save:
            plt.savefig(os.path.join(fig_path, "freetext_barplot.pdf"), format="pdf")
            plt.savefig(os.path.join(fig_path, "freetext_barplot.png"), format="png")

        # Return the figure
        return fig  


    def evaluate_freetext(self, llm_responses: list, model_name: str, run_id: str):
        """
        Evaluate the LLM responses on free-text datasets.
        Currently, FactoolQA, FELM-WK, FactCheck-Bench and FactScore-Bio datasets are included by default.

        Parameters
        ----------
        llm_responses : list
            The LLM responses on the free-text datasets.
        """

        # Set the pipeline for the FreeTextEvaluator
        pipeline = [
            "all_pass_abstain_detector",
            "factool_decontextualizer",
            "factool_evidence_retriever",
            "factool_claim_examiner",
            "factool_post_editor",
            "concat_response_generator"
        ]

        # Initialize the pipeline manually
        self.ofc.init_pipeline_manually(pipeline=pipeline)

        # Get the dataset name and create DataFrame
        dataset = llm_responses[0]['source']
        llm_responses = pd.DataFrame(llm_responses)

        # Save the base_output_path
        self.base_output_path = f"{self.ofc.output_path}/llm_evaluator/{run_id}/{dataset}"

        # Evaluate the LLM responses
        for idx, response in llm_responses.iterrows():
            
            prompt = response['prompt']
            response = response['response']

            # Set the output path
            output_path = f"{self.base_output_path}/{idx}_{md5(prompt.encode()).hexdigest()}"
            if not os.path.exists(output_path):
                os.makedirs(output_path)

            # If the file was already evaluated, skip it
            if (os.path.exists(f"{self.base_output_path}/{idx}_{md5(prompt.encode()).hexdigest()}/evaluation.json")):
                logger.info(f"Skipping the evaluation for prompt {idx} as it was already evaluated.")
                continue

            # TODO: This should work (but it doesn't)
            # self.ofc.init_solver("factool_evidence_retriever", {"path_save_evidence": f"{output_path}/evidence.json"})

            # Evaluate the response
            start = time.time() * 1000
            _result = self.ofc.ResponseEvaluator.evaluate(
                response=response,
                prompt=prompt,
                sample_name=f"llm_evaluator/{run_id}/truth/{dataset}/{idx}"
            )
            end = time.time() * 1000

            # TODO: This is a workaround for the TODO above (move the evidence.json file)
            if os.path.exists("evidence.json"):
                os.rename("evidence.json", f"{output_path}/evidence.json")
            if os.path.exists("evidence_stance.json"):
                os.rename("evidence_stance.json", f"{output_path}/evidence_stance.json")

            # Assess the free-text experiment
            claims = self.assess_freetext(output_path)
            if not claims:
                self.logger.warning(f'Error in assessing experiment for prompt {idx}')
                continue

            # Persist the output
            result = {}
            result["start"] = math.floor(start)
            result["end"] = math.floor(end)
            result["llm"] = model_name
            result["dataset"] = llm_responses["source"][idx]
            result["prompt"] = prompt
            result["claims"] = claims
            result["result"] = _result

            # Save the result
            logger.debug(f"Saving the result for prompt {idx} in {output_path}/evaluation.json")
            with open(f"{output_path}/evaluation.json", "w") as f:
                json.dump(result, f, indent=4)

            logger.info(f"Evaluated the LLM response for prompt {idx} in {end - start} ms.")

        logger.info(f"Finished evaluating the LLM responses for the {dataset} dataset.")

        # Read the outputs
        evaluations = self.read_evaluations()

        # Read the results
        results = self.read_results(evaluations)
       
        return results, evaluations












import openai
import time
import json

from utils.prompt_base import STANCE_DETECTION_PROMPT
from utils.api import chatgpt

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

@Solver.register("chat_gpt_claim_examiner", "evidences", "claim_info")
class ChatGPTClaimExaminer(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.path_save_stance = args.get("path_save_stance", "evidence_stance.json")
        self.num_retries = args.get("num_retries", 3)

    def __call__(self, state: FactCheckerState, *args, **kwargs):

        claim_info = state.get("evidences")

        for key, pair in claim_info.items():
            claim = pair['claim']
            evids = pair['evidence_list']
            if len(evids) == 0:
                claim_info[key]["stances"] = []
                continue

            temp = []
            for i, evid in enumerate(evids):
                user_input = STANCE_DETECTION_PROMPT.format(claim, evid["web_page_snippet_manual"])
                for _ in range(self.num_retries):
                    try:
                        stance = chatgpt(user_input)
                        break
                    except openai.OpenAIError as exception:
                        print(f"{exception}. Retrying...")
                        time.sleep(1)
                # print("Claim: {} \n Evidence: {} \n Stance: {}".format(claim, evid, stance))
                evids[i]["stance"] = stance
                temp.append(stance)
            claim_info[key]["stances"] = temp

        # write to json file
        # Serializing json
        json_object = json.dumps(claim_info, indent=4)

        # Writing to sample.json
        with open(self.path_save_stance, "w") as outfile:
            outfile.write(json_object)

        state.set("claim_info", claim_info)
        return True, state



"""Utils for running the agreement gate."""
import os
import time
from typing import Any, Dict, Tuple

import openai

openai.api_key = os.getenv("OPENAI_API_KEY")


def parse_api_response(api_response: str) -> Tuple[bool, str, str]:
    """Extract the agreement gate state and the reasoning from the GPT-3 API response.

    Our prompt returns questions as a string with the format of an ordered list.
    This function parses this response in a list of questions.

    Args:
        api_response: Agreement gate response from GPT-3.
    Returns:
        is_open: Whether the agreement gate is open.
        reason: The reasoning for why the agreement gate is open or closed.
        decision: The decision of the status of the gate in string form.
    """
    api_response = api_response.strip().split("\n")
    if len(api_response) < 2:
        reason = "Failed to parse."
        decision = None
        is_open = False
    else:
        reason = api_response[0]
        decision = api_response[1].split("Therefore:")[-1].strip()
        is_open = "disagrees" not in api_response[1]
    return is_open, reason, decision


def run_agreement_gate(
    claim: str,
    query: str,
    evidence: str,
    model: str,
    prompt: str,
    context: str = None,
    num_retries: int = 5,
) -> Dict[str, Any]:
    """Checks if a provided evidence contradicts the claim given a query.

    Checks if the answer to a query using the claim contradicts the answer using the
    evidence. If so, we open the agreement gate, which means that we allow the editor
    to edit the claim. Otherwise the agreement gate is closed.

    Args:
        claim: Text to check the validity of.
        query: Query to guide the validity check.
        evidence: Evidence to judge the validity of the claim against.
        model: Name of the OpenAI GPT-3 model to use.
        prompt: The prompt template to query GPT-3 with.
        num_retries: Number of times to retry OpenAI call in the event of an API failure.
    Returns:
        gate: A dictionary with the status of the gate and reasoning for decision.
    """
    if context:
        gpt3_input = prompt.format(
            context=context, claim=claim, query=query, evidence=evidence
        ).strip()
    else:
        gpt3_input = prompt.format(claim=claim, query=query, evidence=evidence).strip()

    for _ in range(num_retries):
        try:
            response = openai.completions.create(
                model=model,
                prompt=gpt3_input,
                temperature=0.0,
                max_tokens=256,
                stop=["\n\n"],
                logit_bias={"50256": -100},  # Don't allow <|endoftext|> to be generated
            )
            break
        except openai.OpenAIError as exception:
            print(f"{exception}. Retrying...")
            time.sleep(2)

    is_open, reason, decision = parse_api_response(response.choices[0].text)
    gate = {"is_open": is_open, "reason": reason, "decision": decision}
    return gate



import os
import uuid
import zipfile
import pandas as pd
import streamlit as st
from importlib import resources as pkg_resources

from openfactcheck.app.utils import metric_card

from openfactcheck import OpenFactCheck
from openfactcheck.templates import llm as templates_dir

# Import solver configuration templates
questions_templates_path = str(pkg_resources.files(templates_dir) / "questions.csv")


def evaluate_llm(ofc: OpenFactCheck):
    """
    This function creates a Streamlit app to evaluate the factuality of a LLM.
    """
    # Initialize the LLM Evaluator
    llm_evaluator = ofc.LLMEvaluator

    st.info("Evaluate the factuality of a Large Language Model (LLM) using *FactQA* Dataset.")

    # Display the instructions
    st.write("Download the questions and instructions to evaluate the factuality of a LLM.")

    # Check if the file exists
    if os.path.exists(questions_templates_path):
        # Create a ZIP file in memory
        from io import BytesIO

        memory_file = BytesIO()
        with zipfile.ZipFile(memory_file, "w") as zf:
            # Define the name of the file within the ZIP archive
            zip_path = os.path.basename(questions_templates_path)  # 'questions.csv'
            # Add file to the ZIP file
            zf.write(questions_templates_path, arcname=zip_path)

        # Reset pointer to start of the memory file
        memory_file.seek(0)

        # Create a download button and the file will be downloaded when clicked
        btn = st.download_button(
            label="Download", data=memory_file, file_name="openfactcheck_llm_benchmark.zip", mime="application/zip"
        )
    else:
        st.error("File not found.")

    # Display the instructions
    st.write("Upload the model responses as a JSON file below to evaluate the factuality.")

    # Upload the model output
    uploaded_file = st.file_uploader("Upload", type=["csv"], label_visibility="collapsed")

    # Check if the file is uploaded
    if uploaded_file is None:
        st.info("Please upload a CSV file.")
        return

    # Check if the file is a CSV file
    if uploaded_file.type != "text/csv":
        st.error("Invalid file format. Please upload a CSV file.")
        return

    # Read the CSV file
    uploaded_data = pd.read_csv(uploaded_file)

    # Ask user to select datasets they want to evaluate on
    st.write("Please select the datasets you want to evaluate the LLM on.")
    datasets = st.multiselect(
        "Select datasets",
        ["snowballing", "selfaware", "freshqa", "factoolqa", "felm-wk", "factcheck-bench", "factscore-bio"],
    )

    def update_first_name():
        st.session_state.first_name = st.session_state.input_first_name

    def update_last_name():
        st.session_state.last_name = st.session_state.input_last_name

    def update_email():
        st.session_state.email = st.session_state.input_email

    def update_organization():
        st.session_state.organization = st.session_state.input_organization

    def update_llm_model():
        st.session_state.llm_model = st.session_state.input_llm_model

    def update_include_in_leaderboard():
        st.session_state.include_in_leaderboard = st.session_state.input_include_in_leaderboard

    # Display instructions
    st.write("Please provide the following information to be included in the leaderboard.")

    # Create text inputs to enter the user information
    st.session_state.id = llm_evaluator.run_id
    st.text_input("First Name", key="input_first_name", on_change=update_first_name)
    st.text_input("Last Name", key="input_last_name", on_change=update_last_name)
    st.text_input("Email", key="input_email", on_change=update_email)
    st.text_input("LLM Model Name", key="input_llm_model", on_change=update_llm_model)
    st.text_input("Organization (Optional)", key="input_organization", on_change=update_organization)

    # Create a checkbox to include the user in the leaderboard
    st.checkbox(
        "Please check this box if you want your LLM to be included in the leaderboard.",
        key="input_include_in_leaderboard",
        on_change=update_include_in_leaderboard,
    )

    if st.button("Evaluate LLM"):
        # Display a success message
        st.success("User information saved successfully.")

        # Display an information message
        st.info(f"""Please wait while we evaluate the factuality of the LLM.
You will be able to download the evaluation report shortly, if you can wait. The report will also be delivered to your email address.
                
Please note your ID {st.session_state.id}, This will be used to track your evaluation.
If the report is not available, please contact the administrator and provide your ID.""")

        # Display a waiting message
        with st.status("Evaluating factuality of the LLM...", expanded=True) as status:
            # Evaluate the LLM
            results = llm_evaluator.evaluate(
                model_name=st.session_state.llm_model, input_path=uploaded_data, datasets=datasets, save_report=False
            )

            # Get plots
            st.write("Generating plots...")
            plots = llm_evaluator.generate_plots(save_plots=False)

            # Generate the evaluation report
            st.write("Generating evaluation report...")
            report_path = llm_evaluator.generate_report(
                report_path=f"{llm_evaluator.output_path}/{llm_evaluator.run_id}"
            )

            status.update(label="LLM evaluated...", state="complete", expanded=False)

        # Display the plots
        st.write("### Evaluation Report")

        # If snowballing dataset is selected
        if "snowballing" in datasets:
            st.write("#### Evaluation on Snowballing Dataset")
            col1, col2, col3 = st.columns(3)
            with col1:
                st.pyplot(plots["snowballing"]["barplot"])
            with col2:
                st.pyplot(plots["snowballing"]["cm"])
            with col3:
                pass

        # If selfaware dataset is selected
        if "selfaware" in datasets:
            st.write("#### Evaluation on SelfAware Dataset")
            col1, col2, col3 = st.columns(3)
            with col1:
                st.pyplot(plots["selfaware"]["barplot"])
            with col2:
                st.pyplot(plots["selfaware"]["cm"])
            with col3:
                pass

        # If freshqa dataset is selected
        if "freshqa" in datasets:
            st.write("#### Evaluation on FreshQA Dataset")
            col1, col2, col3 = st.columns(3)
            with col1:
                st.pyplot(plots["freshqa"]["piechart"])
            with col2:
                pass
            with col3:
                pass

        # If any of the free-text datasets are selected
        if any(dataset in ["factoolqa", "felm-wk", "factcheck-bench", "factscore-bio"] for dataset in datasets):
            st.write("#### Evaluation on Free-Text Datasets")
            st.pyplot(plots["freetext"]["barplot"])

        # Generate the evaluation report
        st.write("### Download Evaluation Report")
        st.info("The report will also be sent to your email address.")

        # Load the evaluation report
        if os.path.exists(report_path):
            with open(report_path, "rb") as file:
                report_bytes = file.read()

                # Display the download button
                st.download_button(
                    label="Download", data=report_bytes, file_name="llm_evaluation_report.pdf", mime="application/pdf"
                )
        else:
            st.error("File not found.")



import itertools
from typing import Any, Dict, List

import torch
from sentence_transformers import CrossEncoder

PASSAGE_RANKER = CrossEncoder(
    "cross-encoder/ms-marco-MiniLM-L-6-v2",
    max_length=512,
    device=torch.device("cuda" if torch.cuda.is_available() else "cpu"),
)


def compute_score_matrix(
    questions: List[str], evidences: List[str]
) -> List[List[float]]:
    """Scores the relevance of all evidence against all questions using a CrossEncoder.

    Args:
        questions: A list of unique questions.
        evidences: A list of unique evidences.
    Returns:
        score_matrix: A 2D list list of question X evidence relevance scores.
    """
    score_matrix = []
    for q in questions:
        evidence_scores = PASSAGE_RANKER.predict([(q, e) for e in evidences]).tolist()
        score_matrix.append(evidence_scores)
    return score_matrix


def question_coverage_objective_fn(
    score_matrix: List[List[float]], evidence_indices: List[int]
) -> float:
    """Given (query, evidence) scores and a subset of evidence, return the coverage.

    Given all pairwise query and evidence scores, and a subset of the evidence
    specified by indices, return a value indicating how well this subset of evidence
    covers (i.e., helps answer) all questions.

    Args:
        score_matrix: A 2D list list of question X evidence relevance scores.
        evidence_indicies: A subset of the evidence to to get the coverage score of.
    Returns:
        total: The coverage we would get by using the subset of evidence in
            `evidence_indices` over all questions.
    """
    # Compute sum_{question q} max_{selected evidence e} score(q, e).
    # This encourages all questions to be explained by at least one evidence.
    total = 0.0
    for scores_for_question in score_matrix:
        total += max(scores_for_question[j] for j in evidence_indices)
    return total


def select_evidences(
    example: Dict[str, Any], max_selected: int = 5, prefer_fewer: bool = False
) -> List[Dict[str, Any]]:
    """Selects the set of evidence that maximizes information converage over the claim.

    Args:
        example: The result of running the editing pipeline on one claim.
        max_selected: Maximum number of evidences to select.
        prefer_fewer: If True and the maximum objective value can be achieved by
            fewer evidences than `max_selected`, prefer selecting fewer evidences.
    Returns:
        selected_evidences: Selected evidences that serve as the attribution report.
    """
    questions = sorted(set(example["questions"]))
    evidences = sorted(set(e["text"] for e in example["revisions"][0]["evidences"]))
    num_evidences = len(evidences)
    if not num_evidences:
        return []

    score_matrix = compute_score_matrix(questions, evidences)

    best_combo = tuple()
    best_objective_value = float("-inf")
    max_selected = min(max_selected, num_evidences)
    min_selected = 1 if prefer_fewer else max_selected
    for num_selected in range(min_selected, max_selected + 1):
        for combo in itertools.combinations(range(num_evidences), num_selected):
            objective_value = question_coverage_objective_fn(score_matrix, combo)
            if objective_value > best_objective_value:
                best_combo = combo
                best_objective_value = objective_value

    selected_evidences = [{"text": evidences[idx]} for idx in best_combo]
    return selected_evidences



import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from openfactcheck.utils.logging import get_logger
from openfactcheck.lib import OpenFactCheckConfig
from openfactcheck.solver import SOLVER_REGISTRY, Solver

# Get the logger
logger = get_logger(__name__)

if TYPE_CHECKING:
    from openfactcheck.evaluator.llm import LLMEvaluator
    from openfactcheck.evaluator.response import ResponseEvaluator
    from openfactcheck.evaluator.checker import CheckerEvaluator


class OpenFactCheck:
    """
    Base class for OpenFactCheck that initializes the solvers and pipeline
    with the given configuration.

    Parameters
    ----------
    config : OpenFactCheckConfig
        An instance of OpenFactCheckConfig containing the configuration
        settings for OpenFactCheck.

    Attributes
    ----------
    logger : Logger
        An instance of the logger to log messages.
    config : OpenFactCheckConfig
        An instance of OpenFactCheckConfig containing the configuration
        settings for OpenFactCheck.
    solver_configs : dict
        A dictionary containing the configuration settings for the solvers.
    pipeline : list
        A list of solvers to be included in the pipeline.
    output_path : str
        The path to the output directory where the results will be stored.

    Methods
    -------
    load_solvers(solver_paths)
        Load solvers from the given paths.
    list_solvers()
        List all registered solvers.
    list_claimprocessors()
        List all registered claim processors.
    list_retrievers()
        List all registered retrievers.
    list_verifiers()
        List all registered verifiers.
    init_solver(solver_name, args)
        Initialize a solver with the given configuration.
    init_solvers()
        Initialize all registered solvers.
    init_pipeline()
        Initialize the pipeline with the given configuration.
    init_pipeline_manually(pipeline)
        Initialize the pipeline with the given configuration.
    persist_output(state, idx, solver_name, cont, sample_name=0)
        Persist the output of the solver.
    read_output(sample_name)
        Read the output file for the given sample.
    remove_output(sample_name)
        Remove the output file for the given sample.
    __call__(response, question, callback_fun, **kwargs)
        Evaluate the response using the pipeline.

    Examples
    --------
    >>> config = OpenFactCheckConfig("config.json")
    >>> ofc = OpenFactCheck(config)
    >>> response, sample_name = ofc("This is a sample response.")
    >>> output = ofc.read_output(sample_name)
    >>> ofc.remove_output(sample_name)
    """

    def __init__(self, config: OpenFactCheckConfig):
        """
        Initialize OpenFactCheck with the given configuration.

        Parameters
        ----------
        config : OpenFactCheckConfig
            An instance of OpenFactCheckConfig containing the configuration
            settings for OpenFactCheck.
        """
        self.logger = logger
        self.config = config

        # Initialize attributes
        self.solver_configs = self.config.solver_configs
        self.pipeline = self.config.pipeline
        self.output_path = os.path.abspath(self.config.output_path)

        # Load and register solvers
        self.load_solvers(self.config.solver_paths)
        self.logger.info(f"Loaded solvers: {list(self.list_solvers().keys())}")

        # Initialize the pipeline
        self.init_pipeline()

    @property
    def LLMEvaluator(self) -> "LLMEvaluator":
        """
        Return the LLM Evaluator
        """
        from openfactcheck.evaluator.llm import LLMEvaluator

        return LLMEvaluator(self)

    @property
    def FactCheckerEvaluator(self) -> "CheckerEvaluator":
        """
        Return the FactChecker Evaluator
        """
        from openfactcheck.evaluator.checker import CheckerEvaluator

        return CheckerEvaluator(self)

    @property
    def ResponseEvaluator(self) -> "ResponseEvaluator":
        """
        Return the LLM Response Evaluator
        """
        from openfactcheck.evaluator.response import ResponseEvaluator

        return ResponseEvaluator(self)

    @staticmethod
    def load_solvers(solver_paths: dict):
        """
        Load solvers from the given paths
        """
        for key, value in solver_paths.items():
            if key == "default":
                for solver_path in value:
                    abs_path = Path(solver_path).resolve()
                    if abs_path.is_dir():
                        sys.path.append(str(abs_path.parent))
                        Solver.load(
                            str(abs_path), f"{abs_path.parent.parent.name}.{abs_path.parent.name}.{abs_path.name}"
                        )
            else:
                for solver_path in value:
                    abs_path = Path(solver_path).resolve()
                    if abs_path.is_dir():
                        sys.path.append(str(abs_path.parent))
                        Solver.load(str(abs_path), abs_path.name)

    @staticmethod
    def list_solvers():
        """
        List all registered solvers
        """
        return SOLVER_REGISTRY

    @staticmethod
    def list_claimprocessors():
        """
        List all registered claim processors
        """
        # Get all claim processors
        claimprocessors = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "claimprocessor" in solver:
                claimprocessors[solver] = value

        return claimprocessors

    @staticmethod
    def list_retrievers():
        """
        List all registered retrievers
        """
        # Get all retrievers
        retrievers = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "retriever" in solver:
                if "evidence" not in solver:  # Exclude evidence retrievers
                    retrievers[solver] = value

        return retrievers

    @staticmethod
    def list_verifiers():
        """
        List all registered verifiers
        """
        # Get all verifiers
        verifiers = {}
        for solver, value in SOLVER_REGISTRY.items():
            if "verifier" in solver:
                verifiers[solver] = value

        return verifiers

    def init_solver(self, solver_name, args):
        """
        Initialize a solver with the given configuration
        """

        # Check if the solver is registered
        if solver_name not in SOLVER_REGISTRY:
            logger.error(f"{solver_name} not in SOLVER_REGISTRY")
            raise RuntimeError(f"{solver_name} not in SOLVER_REGISTRY")

        # Initialize the solver
        solver_cls = SOLVER_REGISTRY[solver_name]
        for key, value in args.items():
            setattr(solver_cls, key, value)

        logger.info(f"Solver {solver_cls(args)} initialized")

        return solver_cls(args), solver_cls.input_name, solver_cls.output_name

    def init_solvers(self):
        """
        Initialize all registered solvers
        """
        solvers = {}
        for k, v in self.solver_configs.items():
            solver, input_name, output_name = self.init_solver(k, v)
            solvers[k] = (solver, input_name, output_name)
        return solvers

    def init_pipeline(self):
        """
        Initialize the pipeline with the given configuration
        """
        self.pipeline = {}
        for required_solver in self.config.pipeline:
            if required_solver not in self.solver_configs:
                logger.error(f"{required_solver} not in solvers config")
                raise RuntimeError(f"{required_solver} not in solvers config")
            solver, input_name, output_name = self.init_solver(required_solver, self.solver_configs[required_solver])
            self.pipeline[required_solver] = (solver, input_name, output_name)

        self.logger.info("-------------- OpenFactCheck Initialized ----------------")
        self.logger.info("Pipeline:")
        for idx, (name, (solver, iname, oname)) in enumerate(self.pipeline.items()):
            self.logger.info(f"{idx}-{name} ({iname} -> {oname})")
        self.logger.info("---------------------------------------------------------")

    def init_pipeline_manually(self, pipeline: list):
        """
        Initialize the pipeline with the given configuration

        Parameters
        ----------
        pipeline : list
            A list of solvers to be included in the pipeline
        """
        self.pipeline = {}
        for required_solver in pipeline:
            if required_solver not in self.solver_configs:
                raise RuntimeError(f"{required_solver} not in solvers config")
            solver, input_name, output_name = self.init_solver(required_solver, self.solver_configs[required_solver])
            self.pipeline[required_solver] = (solver, input_name, output_name)

        self.logger.info("-------------- OpenFactCheck Initialized ----------------")
        self.logger.info("Pipeline:")
        for idx, (name, (solver, iname, oname)) in enumerate(self.pipeline.items()):
            self.logger.info(f"{idx}-{name} ({iname} -> {oname})")
        self.logger.info("---------------------------------------------------------")



import Levenshtein

from .rarr_utils import agreement_gate, editor, evidence_selection
from .prompts import rarr_prompts

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver


@Solver.register("rarr_editor", "claims_with_evidences", "revised_claims")
class RARREditor(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("model", "gpt-4o-instruct")
        # self.model = args.get("model", "gpt-4o-instruct")
        self.max_evidences_per_question = args.get("max_evidences_per_question", 1)
        self.max_edit_ratio = args.get("max_edit_ratio", 100)
        self.output_claim_only = args.get("output_claim_only", False)

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        final_result = {}
        for claim, contents in claims.items():
            context = contents.get("context", None)
            evidences = contents.get("evidences", [])[: self.max_evidences_per_question]
            agreement_gates = []
            revision_steps = []
            claim_for_iterative_revision = claim
            for evidence in evidences:
                gate = agreement_gate.run_agreement_gate(
                    claim=claim_for_iterative_revision,
                    context=context,
                    query=evidence["query"],
                    evidence=evidence["text"],
                    model=self.model,
                    prompt=rarr_prompts.CONTEXTUAL_AGREEMENT_GATE_PROMPT
                    if context
                    else rarr_prompts.AGREEMENT_GATE_PROMPT,
                )
                agreement_gates.append(gate)

                if gate["is_open"]:
                    edited_claim = editor.run_rarr_editor(
                        claim=claim_for_iterative_revision,
                        context=context,
                        query=evidence["query"],
                        evidence=evidence["text"],
                        model=self.model,
                        prompt=rarr_prompts.CONTEXTUAL_EDITOR_PROMPT if context else rarr_prompts.EDITOR_PROMPT,
                    )["text"]
                    if Levenshtein.distance(claim, edited_claim) / len(claim) <= self.max_edit_ratio:
                        claim_for_iterative_revision = edited_claim
                revision_steps.append({"text": claim_for_iterative_revision})
            result = {
                "context": context,
                "text": claim,
                "questions": contents["questions"],
                "evidences_for_questions": evidences,
                "revisions": [
                    {
                        "original_text": claim,
                        "revised_text": revision_steps[-1]["text"],
                        "evidences": evidences,
                        "agreement_gates": agreement_gates,
                        "revision_steps": revision_steps,
                    }
                ],
            }
            selected_evidences = evidence_selection.select_evidences(result)
            result["selected_evidences"] = selected_evidences
            final_result[claim] = result["revisions"][0]["revised_text"] if self.output_claim_only else result
        state.set(self.output_name, final_result)
        return True, state



import os
import argparse
import streamlit as st
from streamlit_option_menu import option_menu

from openfactcheck import OpenFactCheck
from openfactcheck.lib import OpenFactCheckConfig
from openfactcheck.app.sidebar import sidebar
from openfactcheck.app.dialogs import get_secrets
from openfactcheck.app.evaluate_response import evaluate_response
from openfactcheck.app.evaluate_llm import evaluate_llm
from openfactcheck.app.evaluate_factchecker import evaluate_factchecker


def parse_args():
    parser = argparse.ArgumentParser(description="Initialize OpenFactCheck with custom configuration.")

    # Add arguments here, example:
    parser.add_argument("--config-path", type=str, help="Config File Path", default="config.json")

    # Parse arguments from command line
    args = parser.parse_args()
    return args


class App:
    def __init__(self, config_path: str = "config.json"):
        # Set up Dashboard
        st.set_page_config(
            page_title="OpenFactCheck Dashboard",
            page_icon=":bar_chart:",
            layout="wide",
            initial_sidebar_state="collapsed",
        )

        # Get API Keys
        st.session_state.api_keys = False
        # Check if the API keys are already set in the environment variables
        if os.getenv("OPENAI_API_KEY") and os.getenv("SERPER_API_KEY") and os.getenv("SCRAPER_API_KEY"):
            st.session_state.api_keys = True
        else:
            get_secrets()

        # Initialize OpenFactCheck
        @st.cache_resource(show_spinner=False)
        def init(config_path: str = "config.json"):
            # Initialize OpenFactCheckConfig
            config = OpenFactCheckConfig(config_path)
            ofc = OpenFactCheck(config)

            return ofc

        if st.session_state.api_keys:
            self.ofc = init(config_path)
        else:
            self.ofc = None

    def run(self):
        # Set up Sidebar
        sidebar()

        # Remove whitespace from the top of the page and sidebar
        st.markdown(
            """
        <style>
               .block-container {
                    padding-top: 2rem;
                    padding-bottom: 0rem;
                    padding-left: 5rem;
                    padding-right: 5rem;
                }
        </style>
        """,
            unsafe_allow_html=True,
        )

        st.markdown(
            """
            <div style='text-align: center;'>
                <img src='https://raw.githubusercontent.com/hasaniqbal777/OpenFactCheck/main/assets/splash.png' width='400'/>
            </div>
            """,
            unsafe_allow_html=True,
        )
        st.markdown(
            "<h5 style='text-align: center;'>An Open-source Unified Agentic Framework for FactChecking and Factuality Evaluation for LLMs</h5>",
            unsafe_allow_html=True,
        )

        # Selection Menu
        selected = option_menu(
            None,
            ["Evaluate Text", "Evaluate LLM", "Evaluate FactChecker", "Leaderboards", "About"],
            icons=["card-checklist", "check-square", "check2-all", "trophy", "info-circle"],
            menu_icon="cast",
            default_index=0,
            orientation="horizontal",
            # Color scheme
            styles={
                "nav-link-selected": {"background-color": "#005355"},
            },
        )

        # Check if API keys are set
        if not st.session_state.api_keys:
            st.warning("Please provide the necessary API keys to proceed.")
            return

        # Load the selected page
        if selected == "Evaluate Text":
            evaluate_response(self.ofc)
        elif selected == "Evaluate LLM":
            evaluate_llm(self.ofc)
        elif selected == "Evaluate FactChecker":
            evaluate_factchecker(self.ofc)
        # elif selected == "Leaderboards":
        #     leaderboards()
        # else:
        #     about()


if __name__ == "__main__":
    args = parse_args()

    app = App(args.config_path)
    app.run()



import os
import time
from openai import OpenAI
import openai

client = None


def json_fix(output):
    return output.replace("```json\n", "").replace("\n```", "")


def init_client():
    global client
    if client is None:
        if openai.api_key is None and "OPENAI_API_KEY" not in os.environ:
            print("openai_key not presented, delay to initialize.")
            return
        client = OpenAI()


def request(
    user_inputs,
    model,
    system_role,
    temperature=1.0,
    return_all=False,
):
    init_client()

    if type(user_inputs) == str:
        chat_histories = [{"role": "user", "content": user_inputs}]
    elif type(user_inputs) == list:
        if all([type(x) == str for x in user_inputs]):
            chat_histories = [
                {"role": "user" if i % 2 == 0 else "assistant", "content": x} for i, x in enumerate(user_inputs)
            ]
        elif all([type(x) == dict for x in user_inputs]):
            chat_histories = user_inputs
        else:
            raise ValueError("Invalid input for OpenAI API calling")
    else:
        raise ValueError("Invalid input for OpenAI API calling")

    messages = [{"role": "system", "content": system_role}] + chat_histories

    response = client.chat.completions.create(model=model, messages=messages, temperature=temperature)

    if return_all:
        return response
    response_str = ""
    for choice in response.choices:
        response_str += choice.message.content

    response_str = json_fix(response_str)

    return response_str


def gpt(user_inputs, model, system_role, temperature=1.0, num_retries=3, waiting=1):
    response = None
    for _ in range(num_retries):
        try:
            response = request(user_inputs, model, system_role, temperature=temperature)
            break
        except openai.OpenAIError as exception:
            print(f"{exception}. Retrying...")
            time.sleep(waiting)
    return response



import os
import json
import asyncio
import aiohttp

from .chat_api import OpenAIChat
from .prompt import (
    URDU_TO_ENGLISH_TRANSLATION_PROMPT,
    ENGLISH_TO_URDU_TRANSLATION_PROMPT,
)


class GoogleSerperAPIWrapper:
    """Wrapper around the Serper.dev Google Search API.
    You can create a free API key at https://serper.dev.
    To use, you should have the environment variable ``SERPER_API_KEY``
    set with your API key, or pass `serper_api_key` as a named parameter
    to the constructor.
    Example:
        .. code-block:: python
            from langchain import GoogleSerperAPIWrapper
            google_serper = GoogleSerperAPIWrapper()
    """

    def __init__(self, snippet_cnt=10) -> None:
        self.k = snippet_cnt
        self.gl = "us"
        self.hl = "ur"
        self.serper_api_key = os.environ.get("SERPER_API_KEY", None)
        assert (
            self.serper_api_key is not None
        ), "Please set the SERPER_API_KEY environment variable."
        assert (
            self.serper_api_key != ""
        ), "Please set the SERPER_API_KEY environment variable."

        self.gpt_model = os.environ.get("MODEL_NAME", "gpt-4o")
        self.gpt = OpenAIChat(self.gpt_model)
        self.english_to_urdu_translation_prompt = ENGLISH_TO_URDU_TRANSLATION_PROMPT
        self.urdu_to_english_translation_prompt = URDU_TO_ENGLISH_TRANSLATION_PROMPT

    async def _google_serper_search_results(
        self, session, search_term: str, gl: str, hl: str
    ) -> dict:
        headers = {
            "X-API-KEY": self.serper_api_key or "",
            "Content-Type": "application/json",
        }
        params = {"q": search_term, "gl": gl, "hl": hl}
        async with session.post(
            "https://google.serper.dev/search",
            headers=headers,
            params=params,
            raise_for_status=True,
        ) as response:
            return await response.json()

    def _parse_results(self, results):
        snippets = []

        if os.environ.get("SAVE_SERPER_COST", "False") == "True":
            SERPER_COST_PATH = os.environ.get("SERPER_COST_PATH", "serper_cost.jsonl")
            if results.get("credits"):
                credits = results.get("credits")
                with open(SERPER_COST_PATH, "a") as f:
                    f.write(json.dumps({"google_serper_credits": credits}) + "\n")

        if results.get("answerBox"):
            answer_box = results.get("answerBox", {})
            if answer_box.get("answer"):
                element = {"content": answer_box.get("answer"), "source": "None"}
                return [element]
            elif answer_box.get("snippet"):
                element = {
                    "content": answer_box.get("snippet").replace("\n", " "),
                    "source": "None",
                }
                return [element]
            elif answer_box.get("snippetHighlighted"):
                element = {
                    "content": answer_box.get("snippetHighlighted"),
                    "source": "None",
                }
                return [element]

        if results.get("knowledgeGraph"):
            kg = results.get("knowledgeGraph", {})
            title = kg.get("title")
            entity_type = kg.get("type")
            if entity_type:
                element = {"content": f"{title}: {entity_type}", "source": "None"}
                snippets.append(element)
            description = kg.get("description")
            if description:
                element = {"content": description, "source": "None"}
                snippets.append(element)
            for attribute, value in kg.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": "None"}
                snippets.append(element)

        for result in results["organic"][: self.k]:
            if "snippet" in result:
                element = {"content": result["snippet"], "source": result["link"]}
                snippets.append(element)
            for attribute, value in result.get("attributes", {}).items():
                element = {"content": f"{attribute}: {value}", "source": result["link"]}
                snippets.append(element)

        if len(snippets) == 0:
            element = {
                "content": "No good Google Search Result was found",
                "source": "None",
            }
            return [element]

        # keep only the first k snippets
        snippets = snippets[: int(self.k / 2)]

        return snippets

    async def parallel_searches(self, search_queries, gl, hl):
        async with aiohttp.ClientSession() as session:
            tasks = [
                self._google_serper_search_results(session, query, gl, hl)
                for query in search_queries
            ]
            search_results = await asyncio.gather(*tasks, return_exceptions=True)
            return search_results

    def run(self, queries):
        """Run query through GoogleSearch by translating to English first and
        then translating the snippets back to Urdu—no thresholding."""
        # Flatten the nested query lists
        flattened_queries = []
        for sublist in queries:
            if sublist is None:
                sublist = ["None", "None"]
            for item in sublist:
                flattened_queries.append(item)

        # 1) Translate all Urdu queries into English
        messages_to_en = [
            [
                {
                    "role": "system",
                    "content": self.urdu_to_english_translation_prompt["system"],
                },
                {
                    "role": "user",
                    "content": self.urdu_to_english_translation_prompt["user"].format(
                        input=query
                    ),
                },
            ]
            for query in flattened_queries
        ]
        english_queries = self.gpt.run(messages_to_en, str)

        # 2) Perform all searches in English
        results_en = asyncio.run(
            self.parallel_searches(english_queries, gl=self.gl, hl="en")
        )
        parsed_snippets_en = [self._parse_results(r) for r in results_en]

        # 3) Pair up snippets two by two (to match the original logic)
        snippets_pairs = [
            parsed_snippets_en[i] + parsed_snippets_en[i + 1]
            for i in range(0, len(parsed_snippets_en), 2)
        ]

        # 4) Translate each snippet back into Urdu
        final_snippets = []
        for snippet_list in snippets_pairs:
            # build translation prompts for each snippet
            messages_to_ur = [
                [
                    {
                        "role": "system",
                        "content": self.english_to_urdu_translation_prompt["system"],
                    },
                    {
                        "role": "user",
                        "content": self.english_to_urdu_translation_prompt[
                            "user"
                        ].format(input=snip["content"]),
                    },
                ]
                for snip in snippet_list
            ]
            urdu_texts = self.gpt.run(messages_to_ur, str)
            # collect into the same structure
            final_snippets.append(
                [
                    {"content": txt, "source": snippet_list[idx].get("source", "None")}
                    for idx, txt in enumerate(urdu_texts)
                ]
            )

        return final_snippets



"""Utils for searching a query and returning top passages from search results."""
import concurrent.futures
import itertools
import os
import random
from typing import Any, Dict, List, Tuple

import bs4
import requests
import spacy
import torch
from sentence_transformers import CrossEncoder

PASSAGE_RANKER = CrossEncoder(
    "cross-encoder/ms-marco-MiniLM-L-6-v2",
    max_length=512,
    device="cpu",
)
SEARCH_URL = "https://api.bing.microsoft.com/v7.0/search/"
SUBSCRIPTION_KEY = os.getenv("AZURE_SEARCH_KEY")
TOKENIZER = spacy.load("en_core_web_sm", disable=["ner", "tagger", "lemmatizer"])


def chunk_text(
    text: str,
    sentences_per_passage: int,
    filter_sentence_len: int,
    sliding_distance: int = None,
) -> List[str]:
    """Chunks text into passages using a sliding window.

    Args:
        text: Text to chunk into passages.
        sentences_per_passage: Number of sentences for each passage.
        filter_sentence_len: Maximum number of chars of each sentence before being filtered.
        sliding_distance: Sliding distance over the text. Allows the passages to have
            overlap. The sliding distance cannot be greater than the window size.
    Returns:
        passages: Chunked passages from the text.
    """
    if not sliding_distance or sliding_distance > sentences_per_passage:
        sliding_distance = sentences_per_passage
    assert sentences_per_passage > 0 and sliding_distance > 0

    passages = []
    try:
        doc = TOKENIZER(text[:500000])  # Take 500k chars to not break tokenization.
        sents = [
            s.text
            for s in doc.sents
            if len(s.text) <= filter_sentence_len  # Long sents are usually metadata.
        ]
        for idx in range(0, len(sents), sliding_distance):
            passages.append(" ".join(sents[idx : idx + sentences_per_passage]))
    except UnicodeEncodeError as _:  # Sometimes run into Unicode error when tokenizing.
        print("Unicode error when using Spacy. Skipping text.")

    return passages


def is_tag_visible(element: bs4.element) -> bool:
    """Determines if an HTML element is visible.

    Args:
        element: A BeautifulSoup element to check the visiblity of.
    returns:
        Whether the element is visible.
    """
    if element.parent.name in [
        "style",
        "script",
        "head",
        "title",
        "meta",
        "[document]",
    ] or isinstance(element, bs4.element.Comment):
        return False
    return True


def scrape_url(url: str, timeout: float = 3) -> Tuple[str, str]:
    """Scrapes a URL for all text information.

    Args:
        url: URL of webpage to scrape.
        timeout: Timeout of the requests call.
    Returns:
        web_text: The visible text of the scraped URL.
        url: URL input.
    """
    # Scrape the URL
    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()
    except requests.exceptions.RequestException as _:
        return None, url

    # Extract out all text from the tags
    try:
        soup = bs4.BeautifulSoup(response.text, "html.parser")
        texts = soup.findAll(text=True)
        # Filter out invisible text from the page.
        visible_text = filter(is_tag_visible, texts)
    except Exception as _:
        return None, url

    # Returns all the text concatenated as a string.
    web_text = " ".join(t.strip() for t in visible_text).strip()
    # Clean up spacing.
    web_text = " ".join(web_text.split())
    return web_text, url


def search_bing(query: str, timeout: float = 3) -> List[str]:
    """Searches the query using Bing.
    Args:
        query: Search query.
        timeout: Timeout of the requests call.
    Returns:
        search_results: A list of the top URLs relevant to the query.
    """
    headers = {"Ocp-Apim-Subscription-Key": SUBSCRIPTION_KEY}
    params = {"q": query, "textDecorations": True, "textFormat": "HTML"}
    response = requests.get(SEARCH_URL, headers=headers, params=params, timeout=timeout)
    response.raise_for_status()

    response = response.json()
    search_results = [r["url"] for r in response["webPages"]["value"]]
    return search_results


def run_search(
    query: str,
    cached_search_results: List[str] = None,
    max_search_results_per_query: int = 3,
    max_sentences_per_passage: int = 5,
    sliding_distance: int = 1,
    max_passages_per_search_result_to_return: int = 1,
    timeout: float = 3,
    randomize_num_sentences: bool = False,
    filter_sentence_len: int = 250,
    max_passages_per_search_result_to_score: int = 30,
) -> List[Dict[str, Any]]:
    """Searches the query on a search engine and returns the most relevant information.

    Args:
        query: Search query.
        max_search_results_per_query: Maximum number of search results to get return.
        max_sentences_per_passage: Maximum number of sentences for each passage.
        filter_sentence_len: Maximum length of a sentence before being filtered.
        sliding_distance: Sliding distance over the sentences of each search result.
            Used to extract passages.
        max_passages_per_search_result_to_score: Maxinum number of passages to score for
            each search result.
        max_passages_per_search_result_to_return: Maximum number of passages to return
            for each search result.
    Returns:
        retrieved_passages: Top retrieved passages for the search query.
    """
    if cached_search_results is not None:
        search_results = cached_search_results
    else:
        search_results = search_bing(query, timeout=timeout)

    # Scrape search results in parallel
    with concurrent.futures.ThreadPoolExecutor() as e:
        scraped_results = e.map(scrape_url, search_results, itertools.repeat(timeout))
    # Remove URLs if we weren't able to scrape anything or if they are a PDF.
    scraped_results = [r for r in scraped_results if r[0] and ".pdf" not in r[1]]

    # Iterate through the scraped results and extract out the most useful passages.
    retrieved_passages = []
    for webtext, url in scraped_results[:max_search_results_per_query]:
        if randomize_num_sentences:
            sents_per_passage = random.randint(1, max_sentences_per_passage)
        else:
            sents_per_passage = max_sentences_per_passage

        # Chunk the extracted text into passages.
        passages = chunk_text(
            text=webtext,
            sentences_per_passage=sents_per_passage,
            filter_sentence_len=filter_sentence_len,
            sliding_distance=sliding_distance,
        )
        passages = passages[:max_passages_per_search_result_to_score]
        if not passages:
            continue

        # Score the passages by relevance to the query using a cross-encoder.
        scores = PASSAGE_RANKER.predict([(query, p) for p in passages]).tolist()
        passage_scores = list(zip(passages, scores))

        # Take the top passages_per_search passages for the current search result.
        passage_scores.sort(key=lambda x: x[1], reverse=True)
        for passage, score in passage_scores[:max_passages_per_search_result_to_return]:
            retrieved_passages.append(
                {
                    "text": passage,
                    "url": url,
                    "query": query,
                    "sents_per_passage": sents_per_passage,
                    "retrieval_score": score,  # Cross-encoder score as retr score
                }
            )

    if retrieved_passages:
        # Sort all retrieved passages by the retrieval score.
        retrieved_passages = sorted(
            retrieved_passages, key=lambda d: d["retrieval_score"], reverse=True
        )

        # Normalize the retreival scores into probabilities
        scores = [r["retrieval_score"] for r in retrieved_passages]
        probs = torch.nn.functional.softmax(torch.Tensor(scores), dim=-1).tolist()
        for prob, passage in zip(probs, retrieved_passages):
            passage["score"] = prob

    return retrieved_passages



import os
import uuid
import zipfile
import pandas as pd
import seaborn as sns
import streamlit as st
import matplotlib.pyplot as plt
from importlib import resources as pkg_resources

from openfactcheck.app.utils import metric_card

from openfactcheck import OpenFactCheck
from openfactcheck.templates import factchecker as templates_dir

# Import solver configuration templates
claims_templates_path = str(pkg_resources.files(templates_dir) / "claims.jsonl")
documents_templates_path = str(pkg_resources.files(templates_dir) / "documents.jsonl")


def evaluate_factchecker(ofc: OpenFactCheck):
    """
    This function creates a Streamlit app to evaluate a Factchecker.
    """

    # Initialize the FactChecker Evaluator
    fc_evaluator = ofc.FactCheckerEvaluator

    st.info("Evaluate the factuality of a FactChecker using the *FactCheckBench*.")

    # Display the instructions
    st.write("Download the benchmark evaluate the factuality of a FactChecker.")

    # Check if the file exists
    if os.path.exists(claims_templates_path) and os.path.exists(documents_templates_path):
        # Create a ZIP file in memory
        from io import BytesIO

        memory_file = BytesIO()
        with zipfile.ZipFile(memory_file, "w") as zf:
            # Define the name of the file within the ZIP archive
            zip_path = os.path.basename(claims_templates_path)  # 'claims.jsonl'
            # Add file to the ZIP file
            zf.write(claims_templates_path, arcname=zip_path)

            # TODO: documents.jsonl functionality is still in development
            # zip_path = os.path.basename(documents_templates_path) # 'documents.jsonl'
            # # Add file to the ZIP file
            # zf.write(documents_templates_path, arcname=zip_path)

        # Reset pointer to start of the memory file
        memory_file.seek(0)

        # Create a download button and the file will be downloaded when clicked
        btn = st.download_button(
            label="Download",
            data=memory_file,
            file_name="openfactcheck_factchecker_benchmark.zip",
            mime="application/zip",
        )
    else:
        st.error("File not found.")

    # Display the instructions
    st.write("Upload the FactChecker responses as a JSON file below to evaluate the factuality.")

    # Upload the model output
    uploaded_file = st.file_uploader("Upload", type=["csv"], label_visibility="collapsed")

    # Check if the file is uploaded
    if uploaded_file is None:
        st.info("Please upload a CSV file.")
        return

    # Check if the file is a CSV file
    if uploaded_file.type != "text/csv":
        st.error("Invalid file format. Please upload a CSV file.")
        return

    # Read the CSV file
    uploaded_data = pd.read_csv(uploaded_file)

    def update_first_name():
        st.session_state.first_name = st.session_state.input_first_name

    def update_last_name():
        st.session_state.last_name = st.session_state.input_last_name

    def update_email():
        st.session_state.email = st.session_state.input_email

    def update_organization():
        st.session_state.organization = st.session_state.input_organization

    def update_factchecker():
        st.session_state.factchecker = st.session_state.input_factchecker

    def update_include_in_leaderboard():
        st.session_state.include_in_leaderboard = st.session_state.input_include_in_leaderboard

    # Display instructions
    st.write("Please provide the following information to be included in the leaderboard.")

    # Create text inputs to enter the user information
    st.session_state.id = uuid.uuid4().hex
    st.text_input("First Name", key="input_first_name", on_change=update_first_name)
    st.text_input("Last Name", key="input_last_name", on_change=update_last_name)
    st.text_input("Email", key="input_email", on_change=update_email)
    st.text_input("FactChecker Name", key="input_factchecker", on_change=update_factchecker)
    st.text_input("Organization (Optional)", key="input_organization", on_change=update_organization)

    st.checkbox(
        "Please check this box if you want your FactChecker to be included in the leaderboard.",
        key="input_include_in_leaderboard",
        on_change=update_include_in_leaderboard,
    )

    if st.button("Evaluate FactChecker"):
        # Display a success message
        st.success("User information saved successfully.")

        # Display a waiting message
        with st.status("Evaluating factuality of the FactChecker...", expanded=True) as status:
            result = fc_evaluator.evaluate(input_path=uploaded_data, eval_type="claims")
            status.update(label="FactChecker evaluated...", state="complete", expanded=False)

        # Display the evaluation report
        st.write("### Evaluation report:")

        col1, col2 = st.columns(2, gap="large")
        with col1:
            # Create the heatmap
            classes = ["True", "False"]
            fig = plt.figure()
            sns.heatmap(
                fc_evaluator.confusion_matrix,
                annot=True,
                fmt="d",
                cmap="Blues",
                xticklabels=classes,
                yticklabels=classes,
            )
            plt.ylabel("Actual Class")
            plt.xlabel("Predicted Class")
            st.pyplot(fig)
        with col2:
            # Display the metrics
            accuracy = fc_evaluator.results["True_as_positive"]["accuracy"]
            if accuracy > 0.75 and accuracy <= 1:
                # Green background
                metric_card(
                    label="Accuracy", value=f"{accuracy:.2%}", background_color="#D4EDDA", border_left_color="#28A745"
                )
            elif accuracy > 0.25 and accuracy <= 0.75:
                # Yellow background
                metric_card(
                    label="Accuracy", value=f"{accuracy:.2%}", background_color="#FFF3CD", border_left_color="#FFC107"
                )
            else:
                # Red background
                metric_card(
                    label="Accuracy", value=f"{accuracy:.2%}", background_color="#F8D7DA", border_left_color="#DC3545"
                )

            sub_col1, sub_col2, sub_col3 = st.columns(3)
            with sub_col1:
                metric_card(label="Total Time", value=fc_evaluator.results["total_time"])
            with sub_col2:
                metric_card(label="Total Cost", value=fc_evaluator.results["total_cost"])
            with sub_col3:
                metric_card(label="Number of Samples", value=fc_evaluator.results["num_samples"])

            st.text("Report:\n" + fc_evaluator.classification_report)



"""All prompts used for fact-checking subtasks prompting."""

CLAIM_EXTRACTION_PROMPT = {
    "system": "براہ کرم وہ دعویٰ فراہم کریں جس کی آپ حقیقت جانچنا چاہتے ہیں۔",
    "user": """آپ کو ایک ایسا متن دیا گیا ہے جس میں علم کے دعوے شامل ہیں۔ دعویٰ ایک بیان ہے جو کچھ سچ یا جھوٹ ہونے کا دعویٰ کرتا ہے، جس کی تصدیق انسانوں سے کی جا سکتی ہے۔ آپ کا کام یہ ہے کہ آپ دیے گئے متن میں سے ہر دعوے کو درست طریقے سے شناخت اور نکالیں۔ پھر، کسی بھی کورفرنس (ضمیر یا دوسرے حوالہ دینے والے اظہار) کو دعوے کی وضاحت کے لیے حل کریں۔ ہر دعویٰ مختصر (15 الفاظ سے کم) اور خود مختار ہونا چاہیے۔
متن اردو میں دیا گیا ہے اور دعوے اردو میں نکالے جانے چاہئیں۔
آپ کا جواب صرف نیچے دیے گئے فارمیٹ میں ہونا چاہیے۔ اس کے علاوہ کوئی اور اضافی نوٹس یا وضاحت شامل نہ کریں۔
[جواب کا فارمیٹ]:
[
{{
"claim": "یقین دہانی کرائیں کہ دعویٰ 15 الفاظ سے کم ہو اور مکمل خیال فراہم کرے۔ کورفرنس کو دعوے کی وضاحت کے لیے حل کریں",
}},
…
]

یہاں دو مثالیں دی گئی ہیں:
[text]: کرکٹ میچ میں شعیب ملک نے 50 رنز بنائے۔ وہ پاکستان کے بہترین کھلاڑی ہیں۔ شعیب ملک کو اگلے میچ میں شامل کیا جائے گا۔
[response]: [{{"claim": "شعیب ملک نے 50 رنز بنائے"}}, {{"claim": "شعیب ملک پاکستان کے بہترین کھلاڑی ہیں"}}, {{"claim": "شعیب ملک کو اگلے میچ میں شامل کیا جائے گا"}}]

[text]: لاہور میں موسم خوشگوار رہا۔ لوگ پارکوں میں چلنے پھرنے گئے۔ حکام نے کہا کہ کل بارش ہو سکتی ہے۔
[response]: [{{"claim": "لاہور میں موسم خوشگوار رہا"}}, {{"claim": "لوگ پارکوں میں چلنے پھرنے گئے"}}, {{"claim": "حکام نے کہا کہ کل بارش ہو سکتی ہے"}}]

اب یہ مکمل کریں، صرف جواب کی شکل میں، کوئی اور الفاظ نہیں:
[text]: {input}
[response]:
""",
}

QUERY_GENERATION_PROMPT = {
    "system": "آپ ایک سوالات بنانے والا ہیں جو دیے گئے دعوے کو تصدیق کرنے کے لیے موثر اور جامع تلاش کے انجن کے سوالات تیار کرتا ہے۔ آپ صرف پائیتھون کی فہرست کی شکل میں جواب دیں گے (کسی اور الفاظ میں نہیں!)",
    "user": """آپ ایک سوالات بنانے والے ہیں جو صارفین کو دیے گئے دعوے کو تلاش کے انجن کے ذریعے تصدیق کرنے میں مدد کرتے ہیں۔ آپ کا بنیادی کام دو موثر اور شک انگیز تلاش کے انجن کے سوالات تیار کرنا ہے۔ یہ سوالات صارفین کو دیے گئے دعوے کی حقیقت کو تنقیدی طور پر جانچنے میں مدد فراہم کریں گے۔
سوالات اردو میں ہونے چاہئیں اور سوالات اردو میں بنائے جائیں۔
آپ کو صرف نیچے دیے گئے فارمیٹ میں جواب دینا ہوگا (پائیتھون کی فہرست میں سوالات۔ براہ کرم اس فارمیٹ کی سختی سے پیروی کریں۔ کچھ اور واپس نہ کریں۔ اپنا جواب '[' سے شروع کریں۔
[جواب کا فارمیٹ]: ['سوال1', 'سوال2']

یہاں تین مثالیں ہیں:
دعویٰ: ٹویٹر کے سی ای او بل گیٹس ہیں۔
جواب: ["ٹویٹر کے سی ای او کون ہیں؟", "سی ای او ٹویٹر"]

دعویٰ: مائیکل فیلپس تمام اوقات کے سب سے زیادہ سجاوٹی اولمپین ہیں۔
جواب: ["تمام اوقات کے سب سے زیادہ سجاوٹی اولمپین کون ہیں؟", "مائیکل فیلپس"]

دعویٰ: چیٹ جی پی ٹی کو گوگل نے بنایا ہے۔
جواب: ["چیٹ جی پی ٹی کو کس نے بنایا؟", "چیٹ جی پی ٹی"]

اب یہ مکمل کریں، صرف جواب کی شکل میں، کوئی اور الفاظ نہیں:
دعویٰ: {input}
جواب:
""",
}

VERIFICATION_PROMPT = {
    "system": "آپ ایک شاندار معاون ہیں۔",
    "user": """آپ کو ایک ٹکڑا دیا گیا ہے۔ آپ کا کام یہ ہے کہ آپ یہ شناخت کریں کہ آیا دیے گئے متن میں کوئی حقیقت کی غلطیاں ہیں۔
جب آپ دیے گئے متن کی حقیقت کو پرکھ رہے ہوں، تو آپ ضرورت کے مطابق فراہم کردہ شواہد کا حوالہ دے سکتے ہیں۔ فراہم کردہ شواہد مددگار ہو سکتے ہیں۔ بعض شواہد ایک دوسرے سے متضاد ہو سکتے ہیں۔ آپ کو شواہد کو احتیاط سے استعمال کرنا چاہیے جب آپ دیے گئے متن کی حقیقت کا اندازہ لگائیں۔
جواب ایک ڈکشنری ہونی چاہیے جس میں تین کلیدیں ہوں - "reasoning" (وجہ)، "factuality" (حقیقت) اور "error" (غلطی) اور "correction" (تصحیح)، جو بالترتیب آپ کی وجہ، یہ کہ آیا دیے گئے متن میں کوئی حقیقتی غلطی ہے یا نہیں (Boolean - True یا False)، اور غلطی کی وضاحت، اور تصحیح فراہم کریں۔
وجہ، غلطی اور تصحیح اردو میں ہونی چاہیے۔
یہ ہے دی گئی عبارت
[text]: {claim}
یہ ہیں فراہم کردہ شواہد
[evidences]: {evidence}
آپ کو صرف نیچے دیے گئے فارمیٹ میں جواب دینا چاہیے۔ کچھ اور واپس نہ کریں۔ اپنے جواب کا آغاز '{{' سے کریں۔
[جواب کا فارمیٹ]:
{{
"reasoning": "کیوں دی گئی عبارت حقیقت پر مبنی ہے یا نہیں؟ جب آپ یہ کہتے ہیں کہ کوئی چیز حقیقت پر مبنی نہیں ہے، تو آپ کو اپنے فیصلے کی حمایت کرنے کے لیے متعدد شواہد فراہم کرنے ہوں گے۔",
"error": "اگر عبارت حقیقت پر مبنی ہے تو 'None'، ورنہ غلطی کی وضاحت کریں۔",
"correction": "اگر کوئی غلطی ہو تو تصحیح شدہ عبارت فراہم کریں۔",
"factuality": True اگر دی گئی عبارت حقیقت پر مبنی ہے، ورنہ False۔
}}
""",
}

URDU_TO_ENGLISH_TRANSLATION_PROMPT = {
    "system": "You are a helpful assistant.",
    "user": """You are given a piece of text in Urdu. Your task is to translate it into English. The translation should be accurate and maintain the original meaning of the text. Please ensure that the translation is grammatically correct and coherent in English.
DO NOT RESPOND WITH ANYTHING ELSE. ADDING ANY OTHER EXTRA NOTES THAT VIOLATE THE RESPONSE FORMAT IS BANNED. 

{input}
""",
}

ENGLISH_TO_URDU_TRANSLATION_PROMPT = {
    "system": "You are a helpful assistant.",
    "user": """You are given a piece of text in English. Your task is to translate it into Urdu. The translation should be accurate and maintain the original meaning of the text. Please ensure that the translation is grammatically correct and coherent in Urdu.
DO NOT RESPOND WITH ANYTHING ELSE. ADDING ANY OTHER EXTRA NOTES THAT VIOLATE THE RESPONSE FORMAT IS BANNED.

{input}
""",
}



import os
import re
import time
import pandas as pd
import streamlit as st

from openfactcheck.base import OpenFactCheck
from openfactcheck.app.utils import metric_card


def extract_text(claim):
    """
    Extracts text from a claim that might be a string formatted as a dictionary.
    """
    # Try to extract text using regular expression if claim is a string formatted as a dictionary
    match = re.search(r"'text': '([^']+)'", claim)
    if match:
        return match.group(1)
    return claim  # Return as is if no dictionary format detected


# Create a function to check a LLM response
def evaluate_response(ofc: OpenFactCheck):
    """
    This function creates a Streamlit app to evaluate the factuality of a LLM response.
    """

    # Initialize the response_evaluator
    response_evaluator = ofc.ResponseEvaluator

    # Initialize the solvers
    st.session_state.claimprocessors = {
        "Factool ClaimProcessor": "factool_claimprocessor",
        "FactCheckGPT ClaimProcessor": "factcheckgpt_claimprocessor",
        "UrduFactCheck ClaimProcessor": "urdufactcheck_claimprocessor",
    }
    st.session_state.retrievers = {
        "Factool Retriever": "factool_retriever",
        "FactCheckGPT Retriever": "factcheckgpt_retriever",
        "UrduFactCheck Retriever": "urdufactcheck_retriever",
        "UrduFactCheck Translator Retriever": "urdufactcheck_translator_retriever",
        "UrduFactCheck Thresholded Translator Retriever": "urdufactcheck_thresholded_translator_retriever",
    }
    st.session_state.verifiers = {
        "FactCheckGPT Verifier": "factcheckgpt_verifier",
        "Factool Verifier": "factool_verifier",
        "UrduFactCheck Verifier": "urdufactcheck_verifier",
    }
    st.session_state.claimprocessor = "Factool ClaimProcessor"
    st.session_state.retriever = "Factool Retriever"
    st.session_state.verifier = "FactCheckGPT Verifier"

    st.info(
        "Customize an automatic fact-checker and verify the factuality free-form text. You can select a *claimprocessor*, *retriever*, and *verifier* from the dropdowns below."
    )

    # Dropdown in three columns
    col1, col2, col3 = st.columns(3)
    with col1:
        if "claimprocessor" not in st.session_state:
            claimprocessor_choice = st.selectbox(
                "Select Claim Processor",
                list(st.session_state.claimprocessors.keys()),
                help="Select a claim processor to use for processing claims.",
            )
            st.session_state.claimprocessor = st.session_state.claimprocessors[claimprocessor_choice]
        else:
            claimprocessor_choice = st.selectbox(
                "Select Claim Processor",
                list(st.session_state.claimprocessors.keys()),
                index=list(st.session_state.claimprocessors).index(st.session_state.claimprocessor),
                help="Select a claim processor to use for processing claims.",
            )
            st.session_state.claimprocessor = st.session_state.claimprocessors[claimprocessor_choice]
    with col2:
        if "retriever" not in st.session_state:
            retriever_choice = st.selectbox(
                "Select Retriever",
                list(st.session_state.retrievers.keys()),
                help="Select a retriever to use for retrieving evidences.",
            )
            st.session_state.retriever = st.session_state.retrievers[retriever_choice]
        else:
            retriever_choice = st.selectbox(
                "Select Retriever",
                list(st.session_state.retrievers.keys()),
                index=list(st.session_state.retrievers.keys()).index(st.session_state.retriever),
                help="Select a retriever to use for retrieving evidences.",
            )
            st.session_state.retriever = st.session_state.retrievers[retriever_choice]
    with col3:
        if "verifier" not in st.session_state:
            verifier_choice = st.selectbox(
                "Select Verifier",
                list(st.session_state.verifiers.keys()),
                help="Select a verifier to use for verifying claims.",
            )
            st.session_state.verifier = st.session_state.verifiers[verifier_choice]
        else:
            verifier_choice = st.selectbox(
                "Select Verifier",
                list(st.session_state.verifiers.keys()),
                index=list(st.session_state.verifiers.keys()).index(st.session_state.verifier),
                help="Select a verifier to use for verifying claims.",
            )
            st.session_state.verifier = st.session_state.verifiers[verifier_choice]

    # Your sample responses
    sample_responses = [
        "Elon Musk bought Twitter in 2020 and renamed it to X.",
        "Burj Khalifa is the tallest building in the world and is located in Abu Dhabi. I took a photo in front of it.",
        "برج خلیفہ دنیا کی بلند ترین عمارت ہے اور ابوظہبی میں واقع ہے۔ میں نے اس کے سامنے تصویر کھینچی۔",
    ]

    # Initialize the state for 'input_text' if not already there
    if "input_text" not in st.session_state:
        st.session_state.input_text = ""

    # 3. Define a callback to cycle through responses
    def load_sample():
        current = st.session_state.input_text
        try:
            idx = sample_responses.index(current)
            next_idx = (idx + 1) % len(sample_responses)
        except ValueError:
            next_idx = 0
        st.session_state.input_text = sample_responses[next_idx]

    # 4. Render the textarea, binding it to st.session_state["input_text"]
    st.text_area(
        "Enter LLM response here",
        key="input_text",
        height=150,
        placeholder="Type or paste your free-form text here...",
    )

    # 5. Render the button with on_click=load_sample
    col1, col2 = st.columns([1, 3])
    with col2:
        st.button(
            "Load Sample Response",
            on_click=load_sample,
            use_container_width=True,
            type="secondary",
        )

    with col1:
        # Button to check factuality
        check = st.button("Check Factuality", use_container_width=True, type="primary")

    # Check if the button is clicked
    if check:
        with st.status("Checking factuality...", expanded=True) as status:
            # Configure the pipeline
            st.write("Configuring pipeline...")
            ofc.init_pipeline_manually(
                [st.session_state.claimprocessor, st.session_state.retriever, st.session_state.verifier]
            )
            st.write("Pipeline configured...")

            # Evaluate the response
            st.write("Evaluating response...")

            response = response_evaluator.evaluate_streaming(st.session_state.input_text)
            st.write("Response evaluated...")

            status.update(label="Factuality checked...", state="complete", expanded=False)

        # Display pipeline configuration
        pipeline_str = "&nbsp;&nbsp;&nbsp;┈➤&nbsp;&nbsp;&nbsp;".join(
            [st.session_state.claimprocessor, st.session_state.retriever, st.session_state.verifier]
        )
        st.info(f"""**Pipeline**:&nbsp;&nbsp;&nbsp; \n{pipeline_str}""")

        # Store the final response in the session state
        st.session_state.final_response = None

        col1, col2 = st.columns([3, 1])
        with col1:

            def process_stream(responses):
                """
                Process each response from the stream as a simulated chat output.
                This function yields each word from the formatted text of the response,
                adding a slight delay to simulate typing in a chat.
                """

                for response in responses:
                    if "claimprocessor" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        # Get the number of detected claims
                        detected_claims = output_text.get("claims", [])

                        # Generate formatted text with enumerated claims in Markdown format
                        formatted_text = "### Detected Claims\n"
                        formatted_text += "\n".join(
                            f"{i}. {extract_text(claim)}" for i, claim in enumerate(detected_claims, start=1)
                        )
                        formatted_text += "\n"

                        with col2:
                            metric_card(label="Detected Claims", value=len(detected_claims))

                        # Yield each word with a space and simulate typing by sleeping
                        for word in formatted_text.split(" "):
                            yield word + " "
                            time.sleep(0.01)

                        st.session_state.claimprocessor_flag = True

                    elif "retriever" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        questions = []
                        evidences = []
                        for _, claim_with_evidences in output_text.get("claims_with_evidences", {}).items():
                            for claim_with_evidence in claim_with_evidences:
                                questions.append(claim_with_evidence[0])
                                evidences.append(claim_with_evidence[1])

                        with col2:
                            metric_card(label="Retrieved Evidences", value=len(evidences))

                    elif "verifier" in response["solver_name"]:
                        # Extract response details
                        output_text = response["output"]

                        # Get detail
                        details = output_text.get("detail", None)
                        if details is None:
                            detail_text = "The verifier did not provide any detail. Please use other verifiers for more information."
                        else:
                            detail_text = ""

                            # Apply color to the claim based on factuality
                            claims = 0
                            false_claims = 0
                            true_claims = 0
                            controversial_claims = 0
                            unverified_claims = 0
                            for i, detail in enumerate(details):
                                # Get factuality information
                                factuality = str(detail.get("factuality", None))
                                if factuality is not None:
                                    claim = detail.get("claim", "")
                                    if factuality == "-1" or factuality == "False":
                                        detail_text += f'##### :red[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        false_claims += 1
                                    elif factuality == "1" or factuality == "True":
                                        detail_text += f'##### :green[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        true_claims += 1
                                    elif factuality == "0":
                                        detail_text += f'##### :orange[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        controversial_claims += 1
                                    else:
                                        detail_text += f'##### :purple[{str(i+1) + ". " + extract_text(claim)}]'
                                        detail_text += "\n"
                                        claims += 1
                                        unverified_claims += 1
                                else:
                                    st.error("Factuality not found in the verifier output.")

                                # Add error information
                                if detail.get("error", None) != "None":
                                    detail_text += f"- **Error**: {detail.get('error', '')}"
                                    detail_text += "\n"

                                # Add reasoning information
                                if detail.get("reasoning", None) != "None":
                                    detail_text += f"- **Reasoning**: {detail.get('reasoning', '')}"
                                    detail_text += "\n"

                                # Add correction
                                if detail.get("correction", None) != "":
                                    detail_text += f"- **Correction**: {detail.get('correction', '')}"
                                    detail_text += "\n"

                                # Add evidence
                                if detail.get("evidences", None) != "":
                                    evidence_text = ""
                                    questions_evidences = {}
                                    for evidence in detail.get("evidences", []):
                                        question_evidence = str(evidence[0].split("?")[0]) + "?"
                                        if question_evidence not in questions_evidences:
                                            questions_evidences[question_evidence] = []
                                        questions_evidences[question_evidence].append(evidence[1])
                                    for question, evidences in questions_evidences.items():
                                        evidence_text += f"- **Evidences against Question**: :orange[{question}]"
                                        evidence_text += "\n"
                                        for evidence in evidences:
                                            evidence_text += f"  - {evidence}\n"
                                    detail_text += evidence_text

                        # Generate formatted text with the overall factuality in Markdown format
                        formatted_text = "### Factuality Detail\n"
                        formatted_text += "Factuality of each claim is color-coded (:red[red means false], :green[green means true], :orange[orange means controversial], :violet[violet means unverified]).\n"
                        formatted_text += f"{detail_text}\n"
                        formatted_text += "\n"

                        # Get the number of true and false claims
                        with col2:
                            metric_card(
                                label="Supported Claims",
                                value=true_claims,
                                background_color="#D1ECF1",
                                border_left_color="#17A2B8",
                            )
                            metric_card(
                                label="Conflicted Claims",
                                value=false_claims,
                                background_color="#D1ECF1",
                                border_left_color="#17A2B8",
                            )
                            metric_card(
                                label="Controversial Claims",
                                value=controversial_claims,
                                background_color="#D1ECF1",
                                border_left_color="#17A2B8",
                            )
                            metric_card(
                                label="Unverified Claims",
                                value=unverified_claims,
                                background_color="#D1ECF1",
                                border_left_color="#17A2B8",
                            )

                        # Get overall factuality (label)
                        overall_factuality = output_text.get("label", "Unknown")
                        with col2:
                            with st.container():
                                if overall_factuality:
                                    metric_card(
                                        label="Overall Factuality",
                                        value="True",
                                        background_color="#D4EDDA",
                                        border_left_color="#28A745",
                                    )
                                elif not overall_factuality:
                                    metric_card(
                                        label="Overall Factuality",
                                        value="False",
                                        background_color="#F8D7DA",
                                        border_left_color="#DC3545",
                                    )

                        # Get overall credibility (score)
                        overall_credibility = true_claims / claims if claims > 0 else 0
                        with col2:
                            if overall_credibility > 0.75 and overall_credibility <= 1:
                                # Green background
                                metric_card(
                                    label="Overall Credibility",
                                    value=f"{overall_credibility:.2%}",
                                    background_color="#D4EDDA",
                                    border_left_color="#28A745",
                                )
                            elif overall_credibility > 0.25 and overall_credibility <= 0.75:
                                # Yellow background
                                metric_card(
                                    label="Overall Credibility",
                                    value=f"{overall_credibility:.2%}",
                                    background_color="#FFF3CD",
                                    border_left_color="#FFC107",
                                )
                            else:
                                # Red background
                                metric_card(
                                    label="Overall Credibility",
                                    value=f"{overall_credibility:.2%}",
                                    background_color="#F8D7DA",
                                    border_left_color="#DC3545",
                                )

                        # Yield each word with a space and simulate typing by sleeping
                        for word in formatted_text.split(" "):
                            yield word + " "
                            time.sleep(0.01)

            st.write_stream(process_stream(response))



import re
import bs4
import torch
import spacy
import backoff
import requests
import itertools
import numpy as np
import concurrent.futures
from copy import deepcopy
from openai import RateLimitError
from sentence_transformers import CrossEncoder

from openfactcheck import FactCheckerState, StandardTaskSolver, Solver

from .factcheckgpt_utils.openai_api import gpt
from .factcheckgpt_utils.prompt import QGEN_PROMPT, QGEN_PROMPT_FMT
from .factcheckgpt_utils.data_util import save_txt, save_json


@Solver.register("factcheckgpt_retriever", "claims", "claims_with_evidences")
class FactCheckGPTRetriever(StandardTaskSolver):
    def __init__(self, args):
        super().__init__(args)
        self.model = self.global_config.get("factcheckgpt_model", "gpt-4o")
        self.num_retries = self.global_config.get("num_retries", 3)
        self.tokenizer = spacy.load("en_core_web_sm", disable=["ner", "tagger", "lemmatizer"])
        self.question_duplicate_model = CrossEncoder(
            "navteca/quora-roberta-base", device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        )
        self.passage_ranker = CrossEncoder(
            "cross-encoder/ms-marco-MiniLM-L-6-v2",
            max_length=512,
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        )
        # self.system_role = args.get("system_role", "You are a student full of curiosity")
        self.qgen_system_role = "You are a student full of curiosity"
        self.n_questions = args.get("n_questions", 5)
        self.question_gen_round = args.get("question_gen_round", 1)
        self.qgen_temp = args.get("qgen_temp", 0.7)
        self.search_timeout = args.get("search_timeout", 10)
        self.max_search_results_per_query = args.get("max_search_results_per_query", 5)
        self.max_passages_per_search_result_to_return = args.get("max_passages_per_search_result_to_return", 3)
        self.sentences_per_passage = args.get("sentences_per_passage", 5)
        self.max_passages_per_question = args.get("max_passages_per_question", 5)
        self.max_aggregated_evidences = args.get("max_aggregated_evidences", 5)
        self.question_persist_path = args.get("question_persist_path", "questions.txt")
        self.snippets_persist_path = args.get("snippets_persist_path", "passage.json")

    def __call__(self, state: FactCheckerState, *args, **kwargs):
        claims = state.get(self.input_name)
        claims_with_evidences = {}
        for i, claim in enumerate(claims):
            evidences = self.get_web_evidences_for_claim(claim)
            claims_with_evidences[claim] = [(q, e["text"]) for q, e in evidences["aggregated"]]
        state.set(self.output_name, claims_with_evidences)
        return True, state

    def generate_questions(self, claim, max_loop=5):
        questions = []
        while len(questions) <= 0:
            questions = self.run_question_generation(claim)
            if len(questions) >= 0:
                questions = self.remove_duplicate_questions(questions)
        save_txt(questions, self.question_persist_path)
        return questions

    def retrieve_documents(self, questions):
        snippets = {}
        for question in questions:
            retrieved_passages = self.get_relevant_snippets(question)
            snippets[question] = sorted(retrieved_passages, key=lambda x: x["retrieval_score"], reverse=True)[
                : self.max_passages_per_question
            ]
        save_json(snippets, self.snippets_persist_path)
        return snippets

    def get_web_evidences_for_claim(self, claim):
        evidences = dict()
        evidences["aggregated"] = list()
        questions = self.generate_questions(claim)
        snippets = self.retrieve_documents(questions)
        evidences["question_wise"] = snippets
        total_snippets = sum(list(map(lambda x: len(x), snippets.values())))
        if total_snippets == 0:
            raise RuntimeError("No passages are retrieved, check your network...")
        if total_snippets > self.max_aggregated_evidences:
            while len(evidences["aggregated"]) < self.max_aggregated_evidences:
                for key in evidences["question_wise"]:
                    # Take top evidences for each question
                    if len(evidences["question_wise"][key]) > 0:
                        index = int(len(evidences["aggregated"]) / len(evidences["question_wise"]))
                        evidence = evidences["question_wise"][key][index]
                        evidences["aggregated"].append((key, evidence))
        else:
            evidences["aggregated"] = itertools.chain.from_iterable(
                [[(q, e) for e in es] for q, es in snippets.items()]
            )
        return evidences

    @backoff.on_exception(backoff.expo, RateLimitError)
    def run_question_generation(self, claim):
        questions = set()
        for _ in range(self.question_gen_round):
            user_input = QGEN_PROMPT_FMT.format(claim=claim, n=self.n_questions)
            response = gpt(
                user_input,
                model=self.model,
                system_role=self.qgen_system_role,
                num_retries=self.num_retries,
                temperature=self.qgen_temp,
            )
            try:
                cur_round_questions = set(eval(response))
                questions.update(cur_round_questions)
            except Exception as e:
                print(f"An unexpected error occurred: {e}.")
        questions = list(sorted(questions))
        return questions

    def remove_duplicate_questions(self, all_questions):
        qset = [all_questions[0]]
        for question in all_questions[1:]:
            q_list = [(q, question) for q in qset]
            scores = self.question_duplicate_model.predict(q_list)
            if np.max(scores) < 0.60:
                qset.append(question)
        return qset

    def scrape_url(self, url: str, timeout: float = 3) -> tuple[str, str]:
        """Scrapes a URL for all text information.

        Args:
            url: URL of webpage to scrape.
            timeout: Timeout of the requests call.
        Returns:
            web_text: The visible text of the scraped URL.
            url: URL input.
        """
        # Scrape the URL
        try:
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
        except requests.exceptions.RequestException as _:
            print("URL Error", url)
            return None, url

        # Extract out all text from the tags
        try:
            soup = bs4.BeautifulSoup(response.text, "html.parser")
            texts = soup.findAll(text=True)
            # Filter out invisible text from the page.
            visible_text = filter(self.is_tag_visible, texts)
        except Exception as _:
            print("Parsing Error", response.text)
            return None, url

        # Returns all the text concatenated as a string.
        web_text = " ".join(t.strip() for t in visible_text).strip()
        # Clean up spacing.
        web_text = " ".join(web_text.split())
        return web_text, url

    def is_tag_visible(self, element: bs4.element) -> bool:
        """Determines if an HTML element is visible.

        Args:
            element: A BeautifulSoup element to check the visiblity of.
        returns:
            Whether the element is visible.
        """
        if element.parent.name in [
            "style",
            "script",
            "head",
            "title",
            "meta",
            "[document]",
        ] or isinstance(element, bs4.element.Comment):
            return False
        return True

    def search_google(self, query: str, num_web_pages: int = 10, timeout: int = 6, save_url: str = "") -> list[str]:
        """Searches the query using Google.
        Args:
            query: Search query.
            num_web_pages: the number of web pages to request.
            save_url: path to save returned urls, such as 'urls.txt'
        Returns:
            search_results: A list of the top URLs relevant to the query.
        """
        query = query.replace(" ", "+")

        # set headers: Google returns different web-pages according to agent device
        # desktop user-agent
        USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:65.0) Gecko/20100101 Firefox/65.0"
        # mobile user-agent
        MOBILE_USER_AGENT = "Mozilla/5.0 (Linux; Android 7.0; SM-G930V Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.125 Mobile Safari/537.36"
        headers = {"User-Agent": USER_AGENT}

        # set language
        # set the Google interface language, use &hl=XX
        # set the preferred language of the search results, use &lr=lang_XX
        # set language as en, otherwise it will return many translation web pages to Arabic that can't be opened correctly.
        lang = "en"

        # scrape google results
        urls = []
        for page in range(0, num_web_pages, 10):
            # here page is google search's bottom page meaning, click 2 -> start=10
            # url = "https://www.google.com/search?q={}&start={}".format(query, page)
            url = "https://www.google.com/search?q={}&lr=lang_{}&hl={}&start={}".format(query, lang, lang, page)
            r = requests.get(url, headers=headers, timeout=timeout)
            # collect all urls by regular expression
            # how to do if I just want to have the returned top-k pages?
            urls += re.findall('href="(https?://.*?)"', r.text)

        # set to remove repeated urls
        urls = list(set(urls))

        # save all url into a txt file
        if not save_url == "":
            with open(save_url, "w") as file:
                for url in urls:
                    file.write(url + "\n")
        return urls

    def chunk_text(
        self,
        text: str,
        tokenizer,
        sentences_per_passage: int = 5,
        filter_sentence_len: int = 250,
        sliding_distance: int = 2,
    ) -> list[str]:
        """Chunks text into passages using a sliding window.

        Args:
            text: Text to chunk into passages.
            sentences_per_passage: Number of sentences for each passage.
            filter_sentence_len: Maximum number of chars of each sentence before being filtered.
            sliding_distance: Sliding distance over the text. Allows the passages to have
                overlap. The sliding distance cannot be greater than the window size.
        Returns:
            passages: Chunked passages from the text.
        """
        if not sliding_distance or sliding_distance > sentences_per_passage:
            sliding_distance = sentences_per_passage
        assert sentences_per_passage > 0 and sliding_distance > 0

        passages = []
        try:
            doc = tokenizer(text[:500000])  # Take 500k chars to not break tokenization.
            sents = [
                s.text.replace("\n", " ")
                for s in doc.sents
                if len(s.text) <= filter_sentence_len  # Long sents are usually metadata.
            ]
            for idx in range(0, len(sents), sliding_distance):
                passages.append(
                    (" ".join(sents[idx : idx + sentences_per_passage]), idx, idx + sentences_per_passage - 1)
                )
        except UnicodeEncodeError as _:  # Sometimes run into Unicode error when tokenizing.
            print("Unicode error when using Spacy. Skipping text.")

        return passages

    def get_relevant_snippets(
        self,
        query,
    ):
        search_results = self.search_google(query, timeout=self.search_timeout)

        with concurrent.futures.ThreadPoolExecutor() as e:
            scraped_results = e.map(self.scrape_url, search_results, itertools.repeat(self.search_timeout))
        # Remove URLs if we weren't able to scrape anything or if they are a PDF.
        scraped_results = [r for r in scraped_results if r[0] and ".pdf" not in r[1]]
        # print("Num Bing Search Results: ", len(scraped_results))
        retrieved_passages = list()
        for webtext, url in scraped_results[: self.max_search_results_per_query]:
            passages = self.chunk_text(
                text=webtext, tokenizer=self.tokenizer, sentences_per_passage=self.sentences_per_passage
            )
            if not passages:
                continue

            # Score the passages by relevance to the query using a cross-encoder.
            scores = self.passage_ranker.predict([(query, p[0]) for p in passages]).tolist()
            # Take the top passages_per_search passages for the current search result.
            passage_scores = sorted(zip(passages, scores), reverse=True, key=lambda x: x[1])

            relevant_items = list()
            for passage_item, score in passage_scores:
                overlap = False
                if len(relevant_items) > 0:
                    for item in relevant_items:
                        if passage_item[1] >= item[1] and passage_item[1] <= item[2]:
                            overlap = True
                            break
                        if passage_item[2] >= item[1] and passage_item[2] <= item[2]:
                            overlap = True
                            break

                # Only consider top non-overlapping relevant passages to maximise for information
                if not overlap:
                    relevant_items.append(deepcopy(passage_item))
                    retrieved_passages.append(
                        {
                            "text": passage_item[0],
                            "url": url,
                            "sents_per_passage": self.sentences_per_passage,
                            "retrieval_score": score,  # Cross-encoder score as retr score
                        }
                    )
                if len(relevant_items) >= self.max_passages_per_search_result_to_return:
                    break
        # print("Total snippets extracted: ", len(retrieved_passages))
        return retrieved_passages
